#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
import math

# Namespace: dvg-output
# python library to read diverge output files

DIVERGE_MODEL_MAGIC_NUMBER = ord('D') + ord('M')
DIVERGE_POST_GRID_MAGIC_NUMBER = ord('P')
DIVERGE_POST_PATCH_MAGIC_NUMBER = ord('V')
DIVERGE_POST_TU_MAGIC_NUMBER = ord('T')

# return 1 if A>B, -1 if A<B, and 0 if A=B
def version_compare( A, B ):
    # special cases
    if (A == B):
        return 0
    if (A == "dev"):
        return 1
    if (B == "dev"):
        return -1

    # equalize lengths
    A_list = [int(x) for x in A.lstrip('v').split('.')]
    B_list = [int(x) for x in B.lstrip('v').split('.')]
    vlen = max( len(A_list), len(B_list) )
    if (len(A_list) < vlen):
        A_list += [0] * (vlen - len(A_list))
    if (len(B_list) < vlen):
        B_list += [0] * (vlen - len(B_list))

    # and check for larger version
    for i in range(vlen):
        if A_list[i] > B_list[i]:
            return 1
        elif A_list[i] < B_list[i]:
            return -1
        else:
            pass

    # default: versions are equal
    return 0

# this is due to some weird numpy version checking for the contents of a buffer
def numpy_fromfile( fname, **kwargs ):
    return np.frombuffer( open(fname, "rb").read(), **kwargs )

def qmatrices( data, displ, count, nq ):
    data_d = data[displ:displ+count]
    data_l = data_d.view( dtype=np.longlong )
    def advance( x ):
        nonlocal data_l
        nonlocal data_d
        data_l = data_l[x:]
        data_d = data_d[x:]
    result = []
    for i in range(nq):
        n_k_kp = data_l[0]
        advance(1)
        nb = data_l[0]
        advance(1)
        q = data_l[0]
        advance(1)
        eigen = data_l[0]
        advance(1)
        nv = data_l[0]
        advance(1)
        k_kp_descr = data_l[:n_k_kp].copy()
        advance(n_k_kp)
        if nv == -1 or eigen == 0:
            matrix_d = data_d[:n_k_kp*n_k_kp*nb*nb*nb*nb*2].copy()
            advance(n_k_kp*n_k_kp*nb*nb*nb*nb*2)
            matrix = (matrix_d[:-1:2] + 1j * matrix_d[1::2]).reshape((n_k_kp,nb,nb,n_k_kp,nb,nb))
            values = None
        elif eigen == 1 and nv > 0:
            values_d = data_d[:nv].copy()
            matrix_d = data_d[nv:nv+n_k_kp*nb*nb*2*nv].copy()
            advance(n_k_kp*nb*nb*2*nv + nv)
            matrix = (matrix_d[:-1:2] + 1j * matrix_d[1::2]).reshape((nv,n_k_kp,nb,nb))
            values = values_d
        else:
            print("ERROR. invalid parameters found for packed q-vertex")
        # TODO put eigen/nv info there
        res = (q, k_kp_descr, values, matrix)
        result.append( res )
    return result

def k_ibz_path( byte_ary_slice ):
    if byte_ary_slice.size == 0:
        return None
    else:
        i64 = np.int64
        i8 = i64().itemsize
        n_segments = byte_ary_slice[:i8].view(dtype=i64)[0]
        n_per_segment = byte_ary_slice[i8*1 : i8*(1+n_segments)].view(dtype=i64)
        n_path = byte_ary_slice[i8*(1+n_segments):i8*(2+n_segments)].view(dtype=i64)[0]
        path = byte_ary_slice[i8*(2+n_segments):i8*(2+n_segments+n_path)].view(dtype=i64)
        rest = byte_ary_slice[i8*(2+n_segments+n_path):].view(dtype=np.float64)
        return (n_per_segment, path, rest.reshape((-1,3)))

# Class: diverge_model
# documentation found in docstring, i.e., do the following from within python
# === Python ===
# help(diverge_model)
# ==============
class diverge_model:
    """class to read diverge model files

    name: model name
    version: diverge tag version (serves as file format version)
    dim: dimension
    nk[3]: #kpts
    nkf[3]: #fine kpts

    n_patches: #patches for npatch
    patches: patch indices wrt coarse kmesh
    weights: weight of each patch in BZ integral
    p_count: #refined points per patch
    p_displ: displacement of refined point array for each patch
    p_map: refined indices, indexed by p_count/p_displ
    p_weights: weights of refined indices, indexed by p_count/p_displ

    kc_ibz_path: IBZ path on the coarse momentum mesh as tuple (n_per_segment, path_indices, path_vectors)
    kf_ibz_path: IBZ path on the fine momentum mesh as tuple (n_per_segment, path_indices, path_vectors)

    ibz_path: IBZ path in crystal coordinates

    n_orb: #orbitals
    n_spin: #spins
    SU2: SU2 symmetry?
    lattice: lattice vectors (3,3)
    rlattice: reciprocal lattice vectors (3,3)
    positions: atom positions (n_orb, 3)

    n_sym: #symmetries
    orb_symmetries: orbital symmetries (n_sym, n_orb*n_spin, n_orb*n_spin)
    rs_symmetries: realspace symmetries (n_sym, 3, 3)

    n_hop: #hoppings
    hop: hopping elements (n_hop, custom struct)

    n_vert: #vertex elements
    vert: vertex elements (n_vert, custom struct)

    len_ff_struct: #formfactors
    tu_ff: formfactors (len_ff_struct, custom struct)

    n_vert_chan: #vertex elements in each channel (3)
    data: additional data (bytes)

    kmesh: coarse kmesh (nk[0]*nk[1]*nk[2], 3)
    kfmesh: fine kmesh (nk[0]*nk[1]*nk[2]*nkf[0]*nkf[1]*nkf[2], 3)
    ham: hamiltonain on fine kmesh (kfmesh.shape[0], n_spin*n_orb, n_spin*n_orb)
    U: eigenvectors on fine kmesh (kfmesh.shape[0], n_spin*n_orb, n_spin*n_orb)
    E: eigenvalues on fine kmesh (kfmesh.shape[0], n_spin*n_orb)

    npath: configuration for bandstructure (nonzero after v0.4.x) if -1, used
           kf_ibz_path indices. otherwise, bandstructure array is shaped as
           below
    bandstructure: banstructure along irreducible path, with irreducible path
                   in last three indices. shape usually given as
                   ((ibz_path.size-1)*300+1, n_spin*n_orb + 3) if not specified
                   differently in model output call.
    fatbands: if supplied (using npath == -1), contains array of shape
              (#kf_ibz_path, no_ns, nb) fatbands (nonzero after v0.5.x_dev)
    """

    def _displ_count( self, i_d ):
        return self._f_bytes[self._f_header[i_d]:self._f_header[i_d]+self._f_header[i_d+1]]
    def _head( self, i ):
        return self._f_header[i]
    def _head_displ_count( self, d, c ):
        return self._f_header[d:d+c]

    def __init__(self, fname):
        self._f_header = numpy_fromfile(fname, dtype=np.int64, count=128)

        def check_if_model():
            return (self._f_header[0] == DIVERGE_MODEL_MAGIC_NUMBER)

        def check_numerical_repr():
            checkmask = np.unpackbits(self._f_header[127:].view( dtype=np.uint8 ))
            if checkmask.sum() == 0:
                return True
            else:
                error_bits = np.where(checkmask)[0]
                print( "found errors in bits", error_bits, "(see diverge_model_output.h)" )
                return False

        if not check_if_model():
            self.valid = False
            return
        else:
            self.valid = True

        if not check_numerical_repr():
            self.valid = False

        self.rs_hopping_t = np.dtype(dict(
            names=['R', 'o1', 'o2', 's1', 's2', 't'],
            formats=['3i8', 'i8', 'i8', 'i8', 'i8', 'c16'],
            offsets=self._head_displ_count(102, 6),
            itemsize=self._head(108)))
        self.rs_vertex_t = np.dtype(dict(
            names=['chan', 'R', 'o1', 'o2', 's1', 's2', 's3', 's4', 'V'],
            formats=['b', '3i8', 'i8', 'i8', 'i8', 'i8', 'i8', 'i8', 'c16'],
            offsets=self._head_displ_count(109, 9),
            itemsize=self._head(118)))
        self.tu_formfactor_t = np.dtype(dict(
            names=['R', 'ofrom', 'oto', 'd', 'ffidx'],
            formats=['3i8', 'i8', 'i8', 'f8', 'i8'],
            offsets=self._head_displ_count(119, 5),
            itemsize=self._head(124)))

        self._f_bytes = numpy_fromfile(fname, dtype=np.byte)

        self.name = self._displ_count( 1 ).tobytes().decode().rstrip('\x00')
        self.version = self._displ_count( 125 ).tobytes().decode()
        self.dim = self._head(3)
        self.nk = self._head_displ_count(4,3)
        self.nkf = self._head_displ_count(7,3)

        self.n_patches = self._head(10)
        self.patches = self._displ_count( 11 ).view( dtype=np.int64 )
        self.weights = self._displ_count( 13 ).view( dtype=np.float64 )
        self.p_count = self._displ_count( 15 ).view( dtype=np.int64 )
        self.p_displ = self._displ_count( 17 ).view( dtype=np.int64 )
        self.p_map = self._displ_count( 19 ).view( dtype=np.int64 )
        self.p_weights = self._displ_count( 21 ).view( dtype=np.float64 )

        self.ibz_path = self._displ_count( 23 ).view( dtype=np.float64 ).reshape((-1,3))

        self.n_orb = self._head(25)
        self.n_spin = self._head(46)
        self.SU2 = self._head(45)
        self.lattice = self._head_displ_count(26, 9).view( dtype=np.float64 ).reshape((3,3))
        self.rlattice = self._head_displ_count(58, 9).view( dtype=np.float64 ).reshape((3,3))
        self.positions = self._displ_count( 35 ).view( dtype=np.float64 ).reshape((-1,3))

        self.n_sym = self._head(37)
        self.orb_symmetries = self._displ_count( 38 ).view( dtype=np.complex128 ).reshape(
                (-1, self.n_orb*self.n_spin, self.n_orb*self.n_spin))
        self.rs_symmetries = self._displ_count( 40 ).view( dtype=np.float64 ).reshape((-1,3,3))

        self.n_hop = self._head(42)
        self.hop = self._displ_count( 43 ).view( self.rs_hopping_t )

        self.n_vert = self._head(47)
        self.vert = self._displ_count( 48 ).view( self.rs_vertex_t )

        self.len_ff_struct = self._head( 50 )
        self.tu_ff = self._displ_count( 51 ).view( self.tu_formfactor_t )

        self.n_vert_chan = self._head_displ_count( 53, 3 )
        self.data = self._displ_count( 56 )

        self.kmesh = self._displ_count( 80 ).view( np.float64 ).reshape((-1,3))
        self.kfmesh = self._displ_count( 82 ).view( np.float64 ).reshape((-1,3))
        self.ham = self._displ_count( 84 ).view( np.complex128 ).reshape(
                (-1,self.n_spin*self.n_orb,self.n_spin*self.n_orb))
        self.U = self._displ_count( 86 ).view( np.complex128 ).reshape(
                (-1,self.n_spin*self.n_orb,self.n_spin*self.n_orb))
        self.E = self._displ_count( 88 ).view( np.float64 ).reshape((-1,self.n_spin*self.n_orb))

        self.kc_ibz_path = k_ibz_path( self._displ_count( 90 ) )
        self.kf_ibz_path = k_ibz_path( self._displ_count( 92 ) )

        self.npath = self._head(94) # to see what the config actually is
        self.bandstructure = self._displ_count( 100 ).view(
                np.float64 ).reshape((-1,self.n_spin*self.n_orb + 3))

        fatbands = self._displ_count( 98 ).view( np.float64 )
        if fatbands.size == 0:
            self.fatbands = None
        else:
            self.fatbands = fatbands.reshape( (-1, self.n_spin*self.n_orb, self.n_spin*self.n_orb) )

# Class: diverge_post_patch
# documentation found in docstring, i.e., do the following from within python
# === Python ===
# help(diverge_post_patch)
# ==============
class diverge_post_patch:
    '''class to read diverge postprocessing files for $N$-patch FRG

    version: diverge tag version (serves as file format version)

    vertices and loops may be None if not explicitly asked for in the
    postprocessing routines. They all are of shape (np, np, np, nb, nb, nb, nb),
    i.e. refer to the patch indices.

    V: full vertex
    Lp: particle particle loop
    Lm: particle hole loop
    dV: increment of the last step

    for each interaction channel X (X \in \{P,C,D\}), there are the following
    variables:

    X_chan: boolean that tracks whether this channel is included in the output
    X_dV: boolean that tracks whether the increment was used (True) or whether
          the vertex at scale was used (False)
    X_nq: the number of momentum transfer points
    X_qV: the vertx in 'q representation'. A list of length X_nq with each
          element of the list a tuple (q, k_kp_descr, values, vectors).
          q: transfer momentum
          k_kp_descr: array of indices of secondary momenta. Refers to the
                      coarse kmesh. May differ for different q.
          values: eigenvalues (shape: (nv,) with nv the number of eigenvalues
                  requested) may be None if no eigendecomposition was done
          vectors: eigenvectors (shape: (nv, len(k_kp_descr), nb, nb)). if
                   values is None, nv == len(k_kp_descr)*nb*nb and the array
                   should represents the vertex at the specified q as matrix
                   with shape: (len(k_kp_descr), nb, nb, len(k_kp_descr), nb, nb).
    '''
    def __init__(self, fname):
        d_header = numpy_fromfile(fname, dtype=np.longlong, count=128)

        if d_header[0] == DIVERGE_POST_PATCH_MAGIC_NUMBER:
            self.valid = True
        else:
            self.valid = False
            return

        d_data = numpy_fromfile(fname)
        d_data_l = numpy_fromfile(fname, dtype=np.longlong)

        if d_header[2] != 0:
            self.V = d_data[ 0+d_header[1] : d_header[1]+d_header[2] : 2 ] \
               + 1j* d_data[ 1+d_header[1] : d_header[1]+d_header[2] : 2 ]
        else:
            self.V = None
        if d_header[4] != 0:
            self.Lp = d_data[ 0+d_header[3] : d_header[3]+d_header[4] : 2 ] \
                + 1j* d_data[ 1+d_header[3] : d_header[3]+d_header[4] : 2 ]
        else:
            self.Lp = None
        if d_header[6] != 0:
            self.Lm = d_data[ 0+d_header[5] : d_header[5]+d_header[6] : 2 ] \
                + 1j* d_data[ 1+d_header[5] : d_header[5]+d_header[6] : 2 ]
        else:
            self.Lm = None
        if d_header[8] != 0:
            self.dV = d_data[ 0+d_header[7] : d_header[7]+d_header[8] : 2 ] \
                + 1j* d_data[ 1+d_header[7] : d_header[7]+d_header[8] : 2 ]
        else:
            self.dV = None

        idx = 9
        self.P_chan = chr(d_header[idx]); idx = idx+1
        self.P_dV = bool(d_header[idx]); idx = idx+1
        self.P_nq = d_header[idx]; idx = idx+1
        if d_header[idx+1] > 0:
            self.P_qV = qmatrices( d_data, d_header[idx], d_header[idx+1], self.P_nq ); idx = idx+2
        else:
            self.P_qV = None

        idx = 14
        self.C_chan = chr(d_header[idx]); idx = idx+1
        self.C_dV = bool(d_header[idx]); idx = idx+1
        self.C_nq = d_header[idx]; idx = idx+1
        if d_header[idx+1] > 0:
            self.C_qV = qmatrices( d_data, d_header[idx], d_header[idx+1], self.C_nq ); idx = idx+2
        else:
            self.C_qV = None

        idx = 19
        self.D_chan = chr(d_header[idx]); idx = idx+1
        self.D_dV = bool(d_header[idx]); idx = idx+1
        self.D_nq = d_header[idx]; idx = idx+1
        if d_header[idx+1] > 0:
            self.D_qV = qmatrices( d_data, d_header[idx], d_header[idx+1], self.D_nq ); idx = idx+2
        else:
            self.D_qV = None

        self.version = d_data[ d_header[125] : d_header[125]+d_header[126] ].view(
                dtype=np.byte ).tobytes().decode().rstrip('\x00')

# Class: diverge_post_grid
# documentation found in docstring, i.e., do the following from within python
# === Python ===
# help(diverge_post_grid)
# ==============
class diverge_post_grid:
    '''class to read diverge postprocessing files for grid FRG

    version: diverge tag version (serves as file format version)

    nk: number of k points
    nb: number of bands
    nff: number of formfactors
    SU2: SU2 symmetry
    lingap_num_ev: linearized gap equation number of singular values
    lingap_matrix_size: linearized gap equation matrix size

    formfac: formfactors (nff,nk)

    P_susc: P susceptibility (nk,nff,nb,nb,nb,nb)
    P_mf_U: P lingap U matrix (lingap_num_ev,lingap_matrix_size)
    P_mf_V: P lingap V matrix (lingap_num_ev,lingap_matrix_size)
    P_mf_S: P lingap singular values (lingap_num_ev)
    P_mf_EU: P lingap vertex eigenvectors (lingap_num_ev,lingap_matrix_size)
    P_mf_EV: P lingap vertex eigenvalues (lingap_num_ev)
    C_susc: C susceptibility (nk,nff,nb,nb,nb,nb)
    C_mf_U: C lingap U matrix (lingap_num_ev,lingap_matrix_size)
    C_mf_V: C lingap V matrix (lingap_num_ev,lingap_matrix_size)
    C_mf_S: C lingap singular values (lingap_num_ev)
    C_mf_EU: C lingap vertex eigenvectors (lingap_num_ev,lingap_matrix_size)
    C_mf_EV: C lingap vertex eigenvalues (lingap_num_ev)
    D_susc: D susceptibility (nk,nff,nb,nb,nb,nb)
    D_mf_U: D lingap U matrix (lingap_num_ev,lingap_matrix_size)
    D_mf_V: D lingap V matrix (lingap_num_ev,lingap_matrix_size)
    D_mf_S: D lingap singular values (lingap_num_ev)
    D_mf_EU: D lingap vertex eigenvectors (lingap_num_ev,lingap_matrix_size)
    D_mf_EV: D lingap vertex eigenvalues (lingap_num_ev)

    any of the above can be 'None' if it was not contained in simulation
    '''
    def __init__(self, fname):
        d_header = numpy_fromfile(fname, dtype=np.longlong, count=64)

        if d_header[0] == DIVERGE_POST_GRID_MAGIC_NUMBER:
            self.valid = True
        else:
            self.valid = False
            return

        d_data = numpy_fromfile(fname)

        self.version = d_data[d_header[60]:d_header[60]+d_header[61]].view(
                dtype=np.byte ).tobytes().decode().rstrip('\x00')

        self.nk = d_header[1]
        self.nb = d_header[2]
        self.nff = d_header[3]
        self.SU2 = d_header[4]

        self.lingap_num_ev = d_header[19]
        self.lingap_matrix_size = d_header[20]

        self.file_size = d_header[63]

        formfac = d_data[d_header[5]:d_header[5]+d_header[6]].reshape((self.nff,self.nk,2))
        self.formfac = formfac[:,:,0] + 1j * formfac[:,:,1]

        susc_shape = (self.nk,self.nff,self.nb,self.nb,self.nb,self.nb,2)

        _P_susc = d_data[d_header[7]:d_header[7]+d_header[8]]
        if _P_susc.size != 0:
            self.P_susc = _P_susc.reshape(susc_shape)
            self.P_susc = self.P_susc[:,:, :,:,:,:, 0] + 1j * self.P_susc[:,:, :,:,:,:, 1]
        else:
            self.P_susc = None
        _P_mf = d_data[d_header[9]:d_header[9]+d_header[10]]
        if _P_mf.size != 0:
            self.P_mf_U, self.P_mf_V, self.P_mf_S, self.P_mf_EU, self.P_mf_EV = self.unroll_mf_solution( _P_mf )
        else:
            self.P_mf_U = None
            self.P_mf_V = None
            self.P_mf_S = None
            self.P_mf_EU = None
            self.P_mf_EV = None

        _C_susc = d_data[d_header[11]:d_header[11]+d_header[12]]
        if _C_susc.size != 0:
            self.C_susc = _C_susc.reshape(susc_shape)
            self.C_susc = self.C_susc[:,:, :,:,:,:, 0] + 1j * self.C_susc[:,:, :,:,:,:, 1]
        else:
            self.C_susc = None
        _C_mf = d_data[d_header[13]:d_header[13]+d_header[14]]
        if _C_mf.size != 0:
            self.C_mf_U, self.C_mf_V, self.C_mf_S, self.C_mf_EU, self.C_mf_EV = self.unroll_mf_solution( _C_mf )
        else:
            self.C_mf_U = None
            self.C_mf_V = None
            self.C_mf_S = None
            self.C_mf_EU = None
            self.C_mf_EV = None

        _D_susc = d_data[d_header[15]:d_header[15]+d_header[16]]
        if _D_susc.size != 0:
            self.D_susc = _D_susc.reshape(susc_shape)
            self.D_susc = self.D_susc[:,:, :,:,:,:, 0] + 1j * self.D_susc[:,:, :,:,:,:, 1]
        else:
            self.D_susc = None
        _D_mf = d_data[d_header[17]:d_header[17]+d_header[18]]
        if _D_mf.size != 0:
            self.D_mf_U, self.D_mf_V, self.D_mf_S, self.D_mf_EU, self.D_mf_EV = self.unroll_mf_solution( _D_mf )
        else:
            self.D_mf_U = None
            self.D_mf_V = None
            self.D_mf_S = None
            self.D_mf_EU = None
            self.D_mf_EV = None

    def unroll_mf_solution( self, mf_solution ):
        U = mf_solution[:self.lingap_matrix_size * self.lingap_num_ev * 2]
        V = mf_solution[self.lingap_matrix_size * self.lingap_num_ev * 2:self.lingap_matrix_size * self.lingap_num_ev * 4]
        S = mf_solution[self.lingap_matrix_size * self.lingap_num_ev * 4:self.lingap_matrix_size * self.lingap_num_ev * 4 + self.lingap_num_ev]

        # only present due to change in code (now eigenvalues of vertex are
        # included...)
        if mf_solution.size > self.lingap_matrix_size * self.lingap_num_ev * 4 + self.lingap_num_ev:
            EU = mf_solution[self.lingap_matrix_size * self.lingap_num_ev * 4+self.lingap_num_ev: self.lingap_matrix_size*self.lingap_num_ev*6 +   self.lingap_num_ev]
            EV = mf_solution[self.lingap_matrix_size * self.lingap_num_ev * 6+self.lingap_num_ev: self.lingap_matrix_size*self.lingap_num_ev*6 + 2*self.lingap_num_ev]
        else:
            EU = None
            EV = None

        UU = U.reshape((self.lingap_num_ev,self.lingap_matrix_size,2))
        VV = V.reshape((self.lingap_num_ev,self.lingap_matrix_size,2))
        SS = S.reshape((self.lingap_num_ev,))

        if not (EU is None or EV is None):
            EU = EU.reshape((self.lingap_num_ev,self.lingap_matrix_size,2))
            EU = EU[:,:,0] + 1j * EU[:,:,1]
            EV = EV.reshape((self.lingap_num_ev,))
        return UU[:,:,0] + 1j * UU[:,:,1], VV[:,:,0] + 1j * VV[:,:,1], S, EU, EV

# Class: diverge_post_tu
# documentation found in docstring, i.e., do the following from within python
# === Python ===
# help(diverge_post_tu)
# ==============
class diverge_post_tu:
    '''class to read diverge postprocessing files for tu FRG

    version: diverge tag version (serves as file format version)

    n_orb: number of orbitals (not spin)
    n_spin: number of spin degrees of freedom
    nk: number of momentum points
    nktot: number of momentum for Hamiltonian
    nkibz: number of momentum points in the irreducible BZ wedge
    SU2: Exploit SU2 symmetry
    n_orbff: number of orbital+bond combinations
    n_bonds: maximal number of bonds of a single site, not necessarily the same 
            for every site
    mi_to_ofrom: map from n_orbff to (o,b) notation, stores o (n_orbff)
    mi_to_oto: map from n_orbff to (o,b) notation, stores o+bo (n_orbff)
    mi_to_R: map from n_orbff to (o,b) notation, stores the beyond unit cell
            contributiom (n_orbff,3)
    bond_sizes: number of bonds per site o, needed for correct iteration over
            tu_ff stored in model (n_orb)
    bond_offsets: offset to the bonds belonging to site o (n_orb)
    idx_ibz_in_full_mesh: gives the index of the IBZ point in the full PZ (nkibz)

    kmaps_to: symmetry mapping of full BZ to IBZ (nk)
    mi_to_tuffidx: multiindex to tu_ff index mapping; useful for analysis with model.tu_ff (n_orbff)

    VERTEX DIAGONALISATION
    Plen: for each q point the number of stored elements (nkibz)
    Poff: for each q point the offset in the array (nkibz)
    Ptype: for each q point the used diagonalization algorithm (nkibz)
    Pval: stored Eigenvalues (sum(Plen))
    Pvec: stored Eigenvectors (sum(Plen), n_orb*n_bonds*n_spin**2)

    Clen: for each q point the number of stored elements (nkibz)
    Coff: for each q point the offset in the array (nkibz)
    Ctype: for each q point the used diagonalization algorithm (nkibz)
    Cval: stored Eigenvalues (sum(Clen))
    Cvec: stored Eigenvectors (sum(Clen), n_orb*n_bonds*n_spin**2)

    Dlen: for each q point the number of stored elements (nkibz)
    Doff: for each q point the offset in the array (nkibz)
    Dtype: for each q point the used diagonalization algorithm (nkibz)
    Dval: stored Eigenvalues (sum(Dlen))
    Dvec: stored Eigenvectors (sum(Dlen), n_orb*n_bonds*n_spin**2)

    LINGAP solution
    S_sc: singular values SC lingap (n_sing_val)
    U_sc: U SC lingap (n_sing_val, n_orbff*n_spin**2)
    V_sc: V SC lingap (n_sing_val, n_orbff*n_spin**2)

    S_mag: singular values magnetic lingap (n_sing_val)
    U_mag: U magnetic lingap (n_sing_val, n_orbff*n_spin**2)
    V_mag: V magnetic lingap (n_sing_val, n_orbff*n_spin**2)

    S_ch: singular values charge lingap (n_sing_val)
    U_ch: U charge (n_sing_val, n_orbff*n_spin**2)
    V_ch: V charge (n_sing_val, n_orbff*n_spin**2)

    SUSCEPTIBILITY
    Psusc: Pair-Pair susceptibility ([n_orb*n_spin]*4,nkibz)
    Csusc: magnetic/crossed PH susceptibility ([n_orb*n_spin]*4,nkibz)
    Dsusc: charge/direct PH susceptibility ([n_orb*n_spin]*4,nkibz)
    Psuscff: formfactor resolved Pair-Pair susceptibility ([n_orbff*n_spin*n_spin]*2,nkibz)
    Csuscff: formfactor resolved magnetic/crossed PH susceptibility ([n_orbff*n_spin*n_spin]*2,nkibz)
    Dsuscff: formfactor resolved charge/direct PH susceptibility (n_orbff*n_spin*n_spin]*2,nkibz)
    any of the above can be 'None' if it was not contained in simulation

    SELFENERGY
    selfenergy: Self-energy at the critical scale ([n_orb*n_spin]*2,nktot)


    FULL channels
    Pchannel: P channel on full PZ
    Cchannel: C channel on full PZ
    Dchannel: D channel on full PZ
    pploop: non interacting PP-loop channel on full PZ
    phloop: non interacting PH-loop channel on full PZ

    SYMMETRIES
    symm_o2m_len: length of symmetry maps
    symm_o2m_off: offsets of symmetry maps
    symm_o2m_idx_map: index maps
    self.symm_o2m_Ppref: complex prefactor for P
    self.symm_o2m_Cpref: complex prefactor for C
    '''
    def __init__(self, fname):
        d_header = numpy_fromfile(fname, dtype=np.int64, count=128)

        if d_header[0] == DIVERGE_POST_TU_MAGIC_NUMBER and d_header[126] == DIVERGE_POST_TU_MAGIC_NUMBER:
            self.valid = True
        else:
            self.valid = False
            return

        self.n_orb = int(d_header[1])
        self.n_spin = int(d_header[2])
        self.nk = int(d_header[3])
        self.nktot = int(d_header[4])
        self.nkibz = int(d_header[5])
        self.SU2 = int(d_header[6])
        self.n_orbff = int(d_header[7])
        self.n_bonds = int(d_header[8])
        self.n_sym = int(d_header[9])
        n_spin = self.n_spin
        n_bonds = self.n_bonds
        n_orb = self.n_orb
        n_orbff = self.n_orbff

        _f_bytes = numpy_fromfile(fname, dtype=np.byte)
        def get_array( offset_bytes, count, dtype ):
            return _f_bytes[offset_bytes:offset_bytes + count*dtype().itemsize].view( dtype=dtype )
        def get_array_header( iL, dtype ):
            res = get_array( d_header[iL[0]], d_header[iL[0]+1], dtype )
            iL[0] = iL[0] + 2
            return res

        iL = [10] # can't pass by reference, therefore we need a list... python is stupid after all
        self.mi_to_ofrom = get_array_header( iL, np.int64 )
        self.mi_to_oto = get_array_header( iL, np.longlong )
        self.mi_to_R = get_array_header( iL, np.longlong ).reshape((-1,3))
        self.bond_sizes = get_array_header( iL, np.longlong )
        self.bond_offsets = get_array_header( iL, np.longlong )
        self.idx_ibz_in_bz = get_array_header( iL, np.longlong )
        self.Plen = get_array_header( iL, np.longlong )
        self.Poff = get_array_header( iL, np.longlong )
        self.Ptype = get_array_header( iL, np.byte )
        self.Pval = get_array_header( iL, np.float64 )
        self.Pvec = get_array_header( iL, np.complex128 )

        self.Pvec = self.Pvec.reshape((np.sum(self.Plen),n_spin,n_spin,n_orbff))

        self.Clen = get_array_header( iL, np.longlong )
        self.Coff = get_array_header( iL, np.longlong )
        self.Ctype = get_array_header( iL, np.byte )
        self.Cval = get_array_header( iL, np.float64 )
        self.Cvec = get_array_header( iL, np.complex128 )

        self.Cvec = self.Cvec.reshape((np.sum(self.Clen),n_spin,n_spin,n_orbff))

        self.Dlen = get_array_header( iL, np.longlong )
        self.Doff = get_array_header( iL, np.longlong )
        self.Dtype = get_array_header( iL, np.byte )
        self.Dval = get_array_header( iL, np.float64 )
        self.Dvec = get_array_header( iL, np.complex128 )

        self.Dvec = self.Dvec.reshape((np.sum(self.Dlen),n_spin,n_spin,n_orbff))

        self.S_sc = get_array_header( iL, np.float64 )
        self.n_svdP = d_header[iL[0]-1]
        self.U_sc = get_array_header( iL, np.complex128 )
        self.U_sc = self.U_sc.reshape((self.n_svdP,n_spin,n_spin,n_orbff))
        self.V_sc = get_array_header( iL, np.complex128 )
        self.V_sc = self.V_sc.reshape((self.n_svdP,n_spin,n_spin,n_orbff))

        self.S_mag = get_array_header( iL, np.float64 )
        self.n_svdC = d_header[iL[0]-1]
        self.U_mag = get_array_header( iL, np.complex128 )
        self.U_mag = self.U_mag.reshape((self.n_svdC,n_spin,n_spin,n_orbff))
        self.V_mag = get_array_header( iL, np.complex128 )
        self.V_mag = self.V_mag.reshape((self.n_svdC,n_spin,n_spin,n_orbff))

        self.S_charge = get_array_header( iL, np.float64 )
        self.n_svdD = d_header[iL[0]-1]
        self.U_charge = get_array_header( iL, np.complex128 )
        self.U_charge = self.U_charge.reshape((self.n_svdD,n_spin,n_spin,n_orbff))
        self.V_charge = get_array_header( iL, np.complex128 )
        self.V_charge = self.V_charge.reshape((self.n_svdD,n_spin,n_spin,n_orbff))

        susc_shape = (n_spin,n_orb,n_spin,n_orb,n_spin,n_orb,n_spin,n_orb,self.nk)
        self.Psusc = get_array_header( iL, np.complex128 )
        if(d_header[iL[0]-1] > 0):
            self.Psusc = self.Psusc.reshape(susc_shape)
        self.Csusc = get_array_header( iL, np.complex128 )
        if(d_header[iL[0]-1] > 0):
            self.Csusc = self.Csusc.reshape(susc_shape)
        self.Dsusc = get_array_header( iL, np.complex128 )
        if(d_header[iL[0]-1] > 0):
            self.Dsusc = self.Dsusc.reshape(susc_shape)

        susc_shape = (n_spin*n_spin,n_orbff,n_spin*n_spin,n_orbff,self.nk)
        self.Psusc_ff = get_array_header( iL, np.complex128 )
        if(d_header[iL[0]-1] > 0):
            self.Psusc_ff = self.Psusc_ff.reshape(susc_shape)
        self.Csusc_ff = get_array_header( iL, np.complex128 )
        if(d_header[iL[0]-1] > 0):
            self.Csusc_ff = self.Csusc_ff.reshape(susc_shape)
        self.Dsusc_ff = get_array_header( iL, np.complex128 )
        if(d_header[iL[0]-1] > 0):
            self.Dsusc_ff = self.Dsusc_ff.reshape(susc_shape)

        self.selfenergy = get_array_header( iL, np.complex128 )
        if(d_header[iL[0]-1] > 0):
            self.selfenergy = self.selfenergy.reshape(self.nktot,n_spin,n_orb,n_spin,n_orb)

        self.Pchannel = get_array_header( iL, np.complex128 )
        if(d_header[iL[0]-1] > 0):
            self.Pchannel = self.Pchannel.reshape(susc_shape)
        self.Cchannel = get_array_header( iL, np.complex128 )
        if(d_header[iL[0]-1] > 0):
            self.Cchannel = self.Cchannel.reshape(susc_shape)
        self.Dchannel = get_array_header( iL, np.complex128 )
        if(d_header[iL[0]-1] > 0):
            self.Dchannel = self.Dchannel.reshape(susc_shape)
        self.pploop = get_array_header( iL, np.complex128 )
        if(d_header[iL[0]-1] > 0):
            self.pploop = self.pploop.reshape(susc_shape)
        self.phloop = get_array_header( iL, np.complex128 )
        if(d_header[iL[0]-1] > 0):
            self.phloop = self.phloop.reshape(susc_shape)

        symm_shape = (self.n_sym,n_spin,n_spin,n_orbff)
        self.symm_o2m_len = get_array_header( iL, np.longlong )
        if(d_header[iL[0]-1] > 0):
            self.symm_o2m_len = self.symm_o2m_len.reshape(symm_shape)
        self.symm_o2m_off = get_array_header( iL, np.longlong )
        if(d_header[iL[0]-1] > 0):
            self.symm_o2m_off = self.symm_o2m_off.reshape(symm_shape)
        self.symm_o2m_idx_map = get_array_header( iL, np.longlong )
        self.symm_o2m_Ppref = get_array_header( iL, np.complex128 )
        self.symm_o2m_Cpref = get_array_header( iL, np.complex128 )

        self.version = get_array_header( iL, np.byte ).tobytes().decode()

        self.kmaps_to = get_array_header( iL, np.longlong )
        self.mi_to_tuffidx = get_array_header( iL, np.longlong )

# Function: diverge_read
# function to read model files as well as post processing files. returns the
# corresponding class (<diverge_model>, <diverge_post_patch>, or
# <diverge_post_grid>). Makes use of to the file format specifications given in
# <diverge_postprocess_and_write at
# diverge_postprocess_conf_t.diverge_postprocess_and_write> and
# <diverge_model_to_file>; discerning the different file types by their magic
# numbers
#
# Parameters:
# fname - file name to read from (typically ....dvg)
def read( fname ):
    magic = numpy_fromfile( fname, dtype='i8', count=1 )[0]
    if magic == DIVERGE_MODEL_MAGIC_NUMBER:
        return diverge_model( fname )
    elif magic == DIVERGE_POST_PATCH_MAGIC_NUMBER:
        return diverge_post_patch( fname )
    elif magic == DIVERGE_POST_GRID_MAGIC_NUMBER:
        return diverge_post_grid( fname )
    elif magic == DIVERGE_POST_TU_MAGIC_NUMBER:
        return diverge_post_tu( fname )
    else:
        print(f"magic number {magic} ('{chr(magic)}') unknown")
        return None

# Function: bandstructure_bands
# return the bands array (nk,nb) from the band  structure obtained from a
# <diverge_model>. Useful for plotting (for example see <bandstructure_xvals>).
def bandstructure_bands( model ):
    return model.bandstructure[:,:-3]

# Function: bandstructure_kpts
# return the momentum points (nk,3) from the band structure obtained from a
# <diverge_model>. Useful for plotting and post-processing.
def bandstructure_kpts( model ):
    return model.bandstructure[:,-3:]

# Function: bandstructure_xvals
# calculate the differential distance between the momentum points from the band
# structure obtained from a <diverge_model>. Useful for plotting in conjuction
# with <bandstructure_bands> and <bandstructure_ticks>.
# === Python ===
# import diverge.output as do
# import matplotlib.pyplot as plt
# M = do.read("model.dvg")
# b = do.bandstructure_bands(M)
# x = do.bandstructure_xvals(M)
# plt.plot( x, b, color='black' )
# plt.xticks( do.bandstructure_ticks(M) )
# # need to set the labels manually as they are not known to the model
# plt.show()
# ==============
def bandstructure_xvals( model ):
    K = bandstructure_kpts(model)
    vals = np.concatenate( [[0], np.sqrt((np.diff(K,axis=0)**2).sum(axis=1))] )
    if model.npath == -1 and not model.kf_ibz_path is None:
        n_per_segment = model.kf_ibz_path[0]
        offsets = np.cumsum( n_per_segment )
        offset_zero = offsets[n_per_segment == 0]
        vals[offset_zero] = 0.0
    return np.cumsum(vals)

# Function: bandstructure_ticks
# returns the xticks used for band structure plots from a <diverge_model>.
# Useful in conjuction with <bandstructure_xvals>.
def bandstructure_ticks( model ):
    x = bandstructure_xvals( model )
    n_bandstruct = x.shape[0]
    n_ibz_path = model.ibz_path.shape[0]

    if n_ibz_path <= 1:
        return None

    ticks = []

    if model.npath == 0: # we don't know what the call was. trying to guess.
        n_per_tick_float = (n_bandstruct - 1)/(n_ibz_path-1)
        n_per_tick_int = (n_bandstruct - 1)//(n_ibz_path-1)
        if math.isclose(n_per_tick_int, n_per_tick_float):
            ticks += list( x[ np.arange(n_ibz_path) * n_per_tick_int ] )
        else:
            if not model.kf_ibz_path is None:
                n_per_segment = model.kf_ibz_path[0]
                ticks += list( x[np.concatenate( [[0], np.cumsum( n_per_segment ) - 1] )] )

    elif model.npath >= 1:
        ticks += list( x[ np.arange(n_ibz_path) * model.npath ] )
    elif model.npath == -1:
        if not model.kf_ibz_path is None:
            n_per_segment = model.kf_ibz_path[0]
            ticks += list( x[np.concatenate( [[0], np.cumsum( n_per_segment ) - 1] )] )
        else:
            print( "missing kf_ibz_path in model file" )

    if len(ticks) == n_ibz_path:
        return ticks
    else:
        return None



if __name__ == "__main__":
    # mod = read('post.dvg')
    # print(mod.Plen)
    '''
    print("model name:", mod.name)
    import matplotlib.pyplot as plt
    plt.scatter( mod.kmesh[:,0], mod.kmesh[:,1], c=mod.E[:,0], rasterized=True, lw=0, s=1, cmap=plt.cm.gray_r )
    plt.scatter( *mod.kmesh[mod.patches,:2].T, color='k', s=10 )
    def k1p2( k1, k2, nkx, nky ):
        return ((k1 // nky + k2 // nky) % (nkx)) * nky + (k1 % nky + k2 % nky) % (nky)
    for p in range(len(mod.patches)):
        c = mod.p_count[p]
        d = mod.p_displ[p]
        plt.scatter(*mod.kmesh[k1p2(mod.patches[p], mod.p_map[d:d+c], mod.nk[0], mod.nk[1]),:2].T, lw=0, s=2)
    plt.gca().set_aspect('equal')
    plt.show()
    '''



# Namespace: dvg
# import diverge::
# Wrapper for diverge.h, generated with some help of ctypesgen. It is heavily
# based on wrapping C arrays as numpy structured arrays. Therefore, some
# documentation for these (and how to use them from <dvg>) is included <below at
# Structured Arrays>.
#
# The documentation only includes those functions that exist *on top* of the <C
# library at libdivERGe.so>. Everything else is documented <there at
# libdivERGe.so>. For usage examples, see the python examples shipped with
# <diverge.tar.gz at
# https://git.rwth-aachen.de/frg/diverge/-/raw/master/public/releases/v0.2/divERGe.tar.gz>
# (or found in the <examples/ directory of the repo at
# https://git.rwth-aachen.de/frg/diverge/-/tree/master/examples>).
#
# Note::
# on linux, the python library will search for libdivERGe.so in your
# LD_LIBRARY_PATH _with precedence_ over the shipped version (located somewhere
# in your PYTHONPATH). This makes it easy to swap out the backend by pointing
# LD_LIBRARY_PATH to a location with the desired libdivERGe.so:
# === Code ===
# LD_LIBRARY_PATH=$LD_LIBRARY_PATH:/path/to/gpu/version/ python script.py
# ============

from diverge.helpers import *
from diverge.helpers import String, ReturnString, _libs
from ctypes import *

# Function: info
# print information on the shared library
def info():
    try:
        info_str = str(_libs["divERGe"].access['cdecl'])
    except:
        info_str = "libdivERGe.so not found, or unable to use it."
    try:
        mpi_py_eprint( info_str )
    except:
        pass
    return info_str

if not _libs["divERGe"] is None:

    class struct_complex128_t(Structure):
        pass
    struct_complex128_t.__slots__ = [
        'x',
        'y',
    ]
    struct_complex128_t._fields_ = [
        ('x', c_double),
        ('y', c_double),
    ]
    complex128_t = struct_complex128_t
    gf_complex_t = complex128_t
    index_t = c_int64

    # MPI functions
    init_ = _libs["divERGe"].get("diverge_init", "cdecl")
    init_.argtypes = [POINTER(c_int), POINTER(POINTER(POINTER(c_char)))]
    init_.restype = None
    def init( p_argc=None, p_argv=None ):
        init_( p_argc, p_argv )

    finalize = _libs["divERGe"].get("diverge_finalize", "cdecl")
    finalize.argtypes = []
    finalize.restype = None

    embed = _libs["divERGe"].get("diverge_embed", "cdecl")
    embed.argtypes = [c_voidp]
    embed.restype = None

    reset = _libs["divERGe"].get("diverge_reset", "cdecl")
    reset.argtypes = []
    reset.restype = None

    mpi_exit = _libs["divERGe"].get("diverge_mpi_exit", "cdecl")
    mpi_exit.argtypes = [c_int]
    mpi_exit.restype = None

    mpi_wtime = _libs["divERGe"].get("diverge_mpi_wtime", "cdecl")
    mpi_wtime.argtypes = []
    mpi_wtime.restype = c_double

    mpi_get_comm = _libs["divERGe"].get("diverge_mpi_get_comm", "cdecl")
    mpi_get_comm.argtypes = []
    mpi_get_comm.restype = c_voidp

    mpi_distribute = _libs["divERGe"].get("diverge_mpi_distribute", "cdecl")
    mpi_distribute.argtypes = [index_t]
    mpi_distribute.restype = POINTER(index_t)

    mpi_barrier = _libs["divERGe"].get("diverge_mpi_barrier", "cdecl")
    mpi_barrier.argtypes = []
    mpi_barrier.restype = None

    mpi_comm_size = _libs["divERGe"].get("diverge_mpi_comm_size", "cdecl")
    mpi_comm_size.argtypes = []
    mpi_comm_size.restype = c_int

    mpi_comm_rank = _libs["divERGe"].get("diverge_mpi_comm_rank", "cdecl")
    mpi_comm_rank.argtypes = []
    mpi_comm_rank.restype = c_int

    mpi_allreduce_double_max = _libs["divERGe"].get("diverge_mpi_allreduce_double_max", "cdecl")
    mpi_allreduce_double_max.argtypes = [POINTER(None), POINTER(None), c_int]
    mpi_allreduce_double_max.restype = None

    mpi_allreduce_double_sum = _libs["divERGe"].get("diverge_mpi_allreduce_double_sum", "cdecl")
    mpi_allreduce_double_sum.argtypes = [POINTER(None), POINTER(None), c_int]
    mpi_allreduce_double_sum.restype = None

    mpi_allreduce_complex_sum = _libs["divERGe"].get("diverge_mpi_allreduce_complex_sum", "cdecl")
    mpi_allreduce_complex_sum.argtypes = [POINTER(None), POINTER(None), c_int]
    mpi_allreduce_complex_sum.restype = None

    mpi_allgather_index = _libs["divERGe"].get("diverge_mpi_allgather_index", "cdecl")
    mpi_allgather_index.argtypes = [POINTER(None), POINTER(None), c_int]
    mpi_allgather_index.restype = None

    mpi_allgather_double = _libs["divERGe"].get("diverge_mpi_allgather_double", "cdecl")
    mpi_allgather_double.argtypes = [POINTER(None), POINTER(None), c_int]
    mpi_allgather_double.restype = None

    mpi_send_double = _libs["divERGe"].get("diverge_mpi_send_double", "cdecl")
    mpi_send_double.argtypes = [POINTER(None), c_int, c_int, c_int]
    mpi_send_double.restype = None

    mpi_recv_double = _libs["divERGe"].get("diverge_mpi_recv_double", "cdecl")
    mpi_recv_double.argtypes = [POINTER(None), c_int, c_int, c_int]
    mpi_recv_double.restype = None

    mpi_gatherv_cdoub = _libs["divERGe"].get("diverge_mpi_gatherv_cdoub", "cdecl")
    mpi_gatherv_cdoub.argtypes = [POINTER(None), c_int, POINTER(None), POINTER(c_int), POINTER(c_int), c_int]
    mpi_gatherv_cdoub.restype = None

    mpi_write_cdoub_to_file = _libs["divERGe"].get("diverge_mpi_write_cdoub_to_file", "cdecl")
    mpi_write_cdoub_to_file.argtypes = [String, POINTER(None), c_int, c_int]
    mpi_write_cdoub_to_file.restype = None

    mpi_alltoallv_bytes = _libs["divERGe"].get("diverge_mpi_alltoallv_bytes", "cdecl")
    mpi_alltoallv_bytes.argtypes = [POINTER(None), POINTER(index_t), POINTER(index_t), POINTER(None), POINTER(index_t), POINTER(index_t), index_t]
    mpi_alltoallv_bytes.restype = None

    mpi_alltoallv_complex = _libs["divERGe"].get("diverge_mpi_alltoallv_complex", "cdecl")
    mpi_alltoallv_complex.argtypes = [POINTER(complex128_t), POINTER(c_int), POINTER(c_int), POINTER(complex128_t), POINTER(c_int), POINTER(c_int)]
    mpi_alltoallv_complex.restype = None

    mpi_allgatherv = _libs["divERGe"].get("diverge_mpi_allgatherv", "cdecl")
    mpi_allgatherv.argtypes = [POINTER(complex128_t), POINTER(c_int), POINTER(c_int)]
    mpi_allgatherv.restype = None

    mpi_max = _libs["divERGe"].get("diverge_mpi_max", "cdecl")
    mpi_max.argtypes = [POINTER(c_double)]
    mpi_max.restype = c_double

    mpi_bcast_bytes = _libs["divERGe"].get("diverge_mpi_bcast_bytes", "cdecl")
    mpi_bcast_bytes.argtypes = [POINTER(None), c_int, c_int]
    mpi_bcast_bytes.restype = None

    mpi_gather_double = _libs["divERGe"].get("diverge_mpi_gather_double", "cdecl")
    mpi_gather_double.argtypes = [POINTER(c_double), c_int, POINTER(c_double), c_int, c_int]
    mpi_gather_double.restype = None

    # Threading
    omp_num_threads = _libs["divERGe"].get("diverge_omp_num_threads", "cdecl")
    omp_num_threads.argtypes = None
    omp_num_threads.restype = c_int

    force_thread_limit = _libs["divERGe"].get("diverge_force_thread_limit", "cdecl")
    force_thread_limit.argtypes = [c_int]
    force_thread_limit.restype = None

    # MPI logging
    mpi_loglevel_set = _libs["divERGe"].get("mpi_loglevel_set", "cdecl")
    mpi_loglevel_set.argtypes = [c_int]
    mpi_loglevel_set.restype = None

    mpi_log_set_colors = _libs["divERGe"].get("mpi_log_set_colors", "cdecl")
    mpi_log_set_colors.argtypes = [c_int]
    mpi_log_set_colors.restype = None

    mpi_log_get_colors = _libs["divERGe"].get("mpi_log_get_colors", "cdecl")
    mpi_log_get_colors.argtypes = None
    mpi_log_get_colors.restype = c_int

    mpi_log_control = _libs["divERGe"].get("mpi_log_control", "cdecl")
    mpi_log_control.argtypes = [c_int]
    mpi_log_control.restype = None

    mpi_py_print = _libs["divERGe"].get("mpi_py_print", "cdecl")
    mpi_py_print.argtypes = [String]
    mpi_py_print.restype = None

    mpi_py_eprint = _libs["divERGe"].get("mpi_py_eprint", "cdecl")
    mpi_py_eprint.argtypes = [String]
    mpi_py_eprint.restype = None

    mpi_py_print_all = _libs["divERGe"].get("mpi_py_print_all", "cdecl")
    mpi_py_print_all.argtypes = [String]
    mpi_py_print_all.restype = None

    mpi_py_eprint_all = _libs["divERGe"].get("mpi_py_eprint_all", "cdecl")
    mpi_py_eprint_all.argtypes = [String]
    mpi_py_eprint_all.restype = None

    # Compilation Status
    compilation_status = _libs["divERGe"].get("diverge_compilation_status", "cdecl")
    compilation_status.argtypes = []
    compilation_status.restype = None

    compilation_status = _libs["divERGe"].get("diverge_compilation_status", "cdecl")
    compilation_status.argtypes = []
    compilation_status.restype = None

    compilation_status_mpi = _libs["divERGe"].get("diverge_compilation_status_mpi", "cdecl")
    compilation_status_mpi.argtypes = []
    compilation_status_mpi.restype = c_int
    compilation_status_cuda = _libs["divERGe"].get("diverge_compilation_status_cuda", "cdecl")
    compilation_status_cuda.argtypes = [String]
    compilation_status_cuda.restype = c_int
    compilation_status_version = _libs["divERGe"].get("diverge_compilation_status_version", "cdecl")
    compilation_status_version.argtypes = [String]
    compilation_status_version.restype = c_int
    compilation_status_numbers = _libs["divERGe"].get("diverge_compilation_status_numbers", "cdecl")
    compilation_status_numbers.argtypes = []
    compilation_status_numbers.restype = c_int

    # License
    license = _libs["divERGe"].get("diverge_license", "cdecl")
    license.argtypes = []
    license.restype = String

    license_print = _libs["divERGe"].get("diverge_license_print", "cdecl")
    license_print.argtypes = []
    license_print.restype = None

    # Symmetry Generator
    real_harmonics_t = c_int
    orb_s = 0
    orb_pm1 = 1
    orb_p_0 = 2
    orb_p_1 = 3
    orb_py = 1
    orb_pz = 2
    orb_px = 3
    orb_dm2 = 4
    orb_dm1 = 5
    orb_d0 = 6
    orb_d1 = 7
    orb_d2 = 8
    orb_dxy = 4
    orb_dyz = 5
    orb_dz2 = 6
    orb_dxz = 7
    orb_dx2y2 = 8
    orb_fm3 = 9
    orb_fm2 = 10
    orb_fm1 = 11
    orb_f_0 = 12
    orb_f_1 = 13
    orb_f_2 = 14
    orb_f_3 = 15
    orb_gm4 = 16
    orb_gm3 = 17
    orb_gm2 = 18
    orb_gm1 = 19
    orb_g_0 = 20
    orb_g_1 = 21
    orb_g_2 = 22
    orb_g_3 = 23
    orb_g_4 = 24
    MAX_ORBS_PER_SITE = 20
    class struct_sym_op_t(Structure):
        pass
    sym_op_t = struct_sym_op_t
    struct_sym_op_t.__slots__ = [
        'type',
        'normal_vector',
        'angle',
    ]
    struct_sym_op_t._fields_ = [
        ('type', c_char),
        ('normal_vector', c_double * int(3)),
        ('angle', c_double),
    ]

    class struct_site_descr_t(Structure):
        pass
    site_descr_t = struct_site_descr_t
    struct_site_descr_t.__slots__ = [
        'amplitude',
        'function',
        'n_functions',
        'xaxis',
        'yaxis',
    ]
    struct_site_descr_t._fields_ = [
        ('amplitude', complex128_t * int(MAX_ORBS_PER_SITE)),
        ('function', real_harmonics_t * int(MAX_ORBS_PER_SITE)),
        ('n_functions', index_t),
        ('xaxis', c_double * int(3*MAX_ORBS_PER_SITE)),
        ('yaxis', c_double * int(3*MAX_ORBS_PER_SITE)),
    ]

    generate_symm_trafo_ = _libs["divERGe"].get("diverge_generate_symm_trafo", "cdecl")
    generate_symm_trafo_.argtypes = [ index_t, c_voidp, index_t, c_voidp, index_t, c_voidp, c_voidp ]
    generate_symm_trafo_.restype = None

    # Function: generate_symm_trafo
    # python version of the symmetry transformation
    #
    # Parameters:
    # n_spin - number of spins
    # orbs - np.array( ..., dtype=site_descr_t )
    # syms - np.array( ..., dtype=sym_op_t )
    #
    # Returns:
    # rs_trafo - (3,3) double array
    # orb_trafo - (n_spin*orbs.size, n_spin*orbs.size) complex128_t array
    def generate_symm_trafo( n_spin, orbs, syms ):
        rs_trafo = np.zeros( (3,3), dtype=np.float64 )
        orb_trafo = np.zeros( (n_spin*orbs.size, n_spin*orbs.size), dtype=np.complex128 )

        generate_symm_trafo_( n_spin, orbs.ctypes.data, orbs.size,
                             syms.ctypes.data, syms.size, rs_trafo.ctypes.data,
                             orb_trafo.ctypes.data )
        return rs_trafo, orb_trafo

    # Model

    # structures typedefs
    class struct_rs_hopping_t(Structure):
        pass
    rs_hopping_t = struct_rs_hopping_t

    class struct_rs_vertex_t(Structure):
        pass
    rs_vertex_t = struct_rs_vertex_t

    class struct_mom_patching_t(Structure):
        pass
    mom_patching_t = struct_mom_patching_t

    class struct_model_t(Structure):
        pass
    model_t = struct_model_t

    class struct_internals_t(Structure):
        pass
    internals_t = struct_internals_t

    class struct_tu_formfactor_t(Structure):
        pass
    tu_formfactor_t = struct_tu_formfactor_t

    # function pointers for callbacks
    hamiltonian_generator_t = CFUNCTYPE(UNCHECKED(None), POINTER(model_t), POINTER(complex128_t))
    # greensfunction return values
    enum_greensfunc_op_t = c_int
    greensfunc_op_cpu = 0
    greensfunc_op_gpu = (greensfunc_op_cpu + 1)
    greensfunc_op_t = enum_greensfunc_op_t

    greensfunc_generator_t = CFUNCTYPE(UNCHECKED(greensfunc_op_t), POINTER(model_t), complex128_t, POINTER(gf_complex_t))

    channel_vertex_generator_t = CFUNCTYPE(UNCHECKED(c_int), POINTER(model_t), c_char, POINTER(complex128_t))

    full_vertex_generator_t = CFUNCTYPE(UNCHECKED(None), POINTER(model_t), index_t, index_t, index_t, POINTER(complex128_t))

    # structures
    struct_rs_hopping_t.__slots__ = [
        'R',
        'o1',
        'o2',
        's1',
        's2',
        't',
    ]
    struct_rs_hopping_t._fields_ = [
        ('R', index_t * int(3)),
        ('o1', index_t),
        ('o2', index_t),
        ('s1', index_t),
        ('s2', index_t),
        ('t', complex128_t),
    ]

    struct_rs_vertex_t.__slots__ = [
        'chan',
        'R',
        'o1',
        'o2',
        's1',
        's2',
        's3',
        's4',
        'V',
    ]
    struct_rs_vertex_t._fields_ = [
        ('chan', c_char),
        ('R', index_t * int(3)),
        ('o1', index_t),
        ('o2', index_t),
        ('s1', index_t),
        ('s2', index_t),
        ('s3', index_t),
        ('s4', index_t),
        ('V', complex128_t),
    ]

    struct_mom_patching_t.__slots__ = [
        'n_patches',
        'patches',
        'weights',
        'p_count',
        'p_displ',
        'p_map',
        'p_weights',
    ]
    struct_mom_patching_t._fields_ = [
        ('n_patches', index_t),
        ('patches', POINTER(index_t)),
        ('weights', POINTER(c_double)),
        ('p_count', POINTER(index_t)),
        ('p_displ', POINTER(index_t)),
        ('p_map', POINTER(index_t)),
        ('p_weights', POINTER(c_double)),
    ]

    struct_tu_formfactor_t.__slots__ = [
        'R',
        'ofrom',
        'oto',
        'd',
        'ffidx',
    ]
    struct_tu_formfactor_t._fields_ = [
        ('R', index_t * int(3)),
        ('ofrom', index_t),
        ('oto', index_t),
        ('d', c_double),
        ('ffidx', index_t),
    ]

    struct_model_t.__slots__ = [
        'name',
        'nk',
        'nkf',
        'patching',
        'n_ibz_path',
        'ibz_path',
        'n_orb',
        'lattice',
        'positions',
        'n_sym',
        'orb_symmetries',
        'rs_symmetries',
        'n_hop',
        'hop',
        'hfill',
        'SU2',
        'n_spin',
        'n_vert',
        'vert',
        'tu_ff',
        'n_tu_ff',
        'n_vert_chan',
        'vfill',
        'ffill',
        'gfill',
        'gproj',
        'data',
        'nbytes_data',
        'data_destructor',
        'internals',
    ]
    struct_model_t._fields_ = [
        ('name', c_char * int(1024)),
        ('nk', index_t * int(3)),
        ('nkf', index_t * int(3)),
        ('patching', POINTER(struct_mom_patching_t)),
        ('n_ibz_path', index_t),
        ('ibz_path', (c_double * int(3)) * int(32768)),
        ('n_orb', index_t),
        ('lattice', (c_double * int(3)) * int(3)),
        ('positions', (c_double * int(3)) * int(32768)),
        ('n_sym', index_t),
        ('orb_symmetries', c_voidp),
        ('rs_symmetries', ((c_double * int(3)) * int(3)) * int(256)),
        ('n_hop', index_t),
        ('hop', c_voidp),
        ('hfill', hamiltonian_generator_t),
        ('SU2', c_int),
        ('n_spin', index_t),
        ('n_vert', index_t),
        ('vert', c_voidp),
        ('tu_ff', c_voidp),
        ('n_tu_ff', index_t),
        ('n_vert_chan', index_t * int(3)),
        ('vfill', channel_vertex_generator_t),
        ('ffill', full_vertex_generator_t),
        ('gfill', greensfunc_generator_t),
        ('grpoj', greensfunc_generator_t),
        ('data', String),
        ('nbytes_data', index_t),
        ('data_destructor', CFUNCTYPE(UNCHECKED(None), POINTER(c_char))),
        ('internals', POINTER(internals_t)),
    ]

    # function definitions
    diverge_hamilton_generator_default = _libs["divERGe"].get("diverge_hamilton_generator_default", "cdecl")
    diverge_hamilton_generator_default.argtypes = [POINTER(model_t), POINTER(complex128_t)]
    diverge_hamilton_generator_default.restype = None

    diverge_hamilton_generator_add = _libs["divERGe"].get("diverge_hamilton_generator_add", "cdecl")
    diverge_hamilton_generator_add.argtypes = [POINTER(model_t), POINTER(complex128_t)]
    diverge_hamilton_generator_add.restype = None

    diverge_greensfunc_generator_default = _libs["divERGe"].get("diverge_greensfunc_generator_default", "cdecl")
    diverge_greensfunc_generator_default.argtypes = [POINTER(model_t), complex128_t, POINTER(gf_complex_t)]
    diverge_greensfunc_generator_default.restype = greensfunc_op_t

    diverge_channel_vertex_generator_default = _libs["divERGe"].get("diverge_channel_vertex_generator_default", "cdecl")
    diverge_channel_vertex_generator_default.argtypes = [POINTER(model_t), c_char, POINTER(complex128_t)]
    diverge_channel_vertex_generator_default.restype = c_int

    model_init = _libs["divERGe"].get("diverge_model_init", "cdecl")
    model_init.argtypes = []
    model_init.restype = POINTER(model_t)

    read_fplo = _libs["divERGe"].get("diverge_read_fplo", "cdecl")
    read_fplo.argtypes = [String]
    read_fplo.restype = POINTER(model_t)

    read_W90 = _libs["divERGe"].get("diverge_read_W90_C", "cdecl")
    read_W90.argtypes = [c_char_p,index_t,POINTER(index_t),POINTER(index_t)]
    read_W90.restype = c_void_p
    # Function: read_W90_PY
    # returns a numpy array of <rs_hopping_t> with the hoppings from W90 file
    # and its length
    #
    # Parameters:
    # fname - filename to read W90 data from (usually ..._hr.dat)
    # nspin - default value 0 amounts to SU(2) symmetric model. if nspin != 0,
    #         abs(nspin) = (S+1/2)*2 with S the physical spin (i.e., for S=1/2 we
    #         have abs(nspin)=2). the sign determines whether the spin index is the
    #         one which increases memory slowly (negative) for fast (positive)
    def read_W90_PY( fname, nspin ):
        length = index_t()
        n_orb = index_t(0)
        ptr = POINTER(rs_hopping_t)
        ptr = read_W90( fname.encode('utf-8'), nspin, byref(length), byref(n_orb) )
        pts_ary = view_array( ptr, dtype=rs_hopping_t, shape=(np.array(length),) )
        pts_cpy = np.copy( pts_ary )
        mem_free( ptr )
        return pts_cpy

    rs_hopping_to_supercell = _libs["divERGe"].get("rs_hopping_to_supercell", "cdecl")
    rs_hopping_to_supercell.argtypes = [c_voidp, c_voidp]
    rs_hopping_to_supercell.restype = None

    rs_hopping_to_fractcell = _libs["divERGe"].get("rs_hopping_to_fractcell", "cdecl")
    rs_hopping_to_fractcell.argtypes = [c_voidp, c_voidp, c_voidp, c_voidp]
    rs_hopping_to_fractcell.restype = None

    fukui = _libs["divERGe"].get("diverge_fukui", "cdecl")
    fukui.argtypes = [POINTER(model_t), c_voidp, index_t, c_voidp]
    fukui.restype = c_voidp

    kmesh_to_bands = _libs["divERGe"].get("diverge_kmesh_to_bands", "cdecl")
    kmesh_to_bands.argtypes = [POINTER(model_t), POINTER(POINTER(index_t)), POINTER(index_t)]
    kmesh_to_bands.restype = POINTER(index_t)

    kmesh_to_bands_crs = _libs["divERGe"].get("diverge_kmesh_to_bands_crs", "cdecl")
    kmesh_to_bands_crs.argtypes = [POINTER(model_t), POINTER(POINTER(index_t)), POINTER(index_t)]
    kmesh_to_bands_crs.restype = POINTER(index_t)

    model_free = _libs["divERGe"].get("diverge_model_free", "cdecl")
    model_free.argtypes = [POINTER(model_t)]
    model_free.restype = None

    # Memory allocation routines
    mem_alloc_rs_hopping_t = _libs["divERGe"].get("diverge_mem_alloc_rs_hopping_t", "cdecl")
    mem_alloc_rs_hopping_t.argtypes = [index_t]
    mem_alloc_rs_hopping_t.restype = POINTER(rs_hopping_t)

    mem_alloc_rs_vertex_t = _libs["divERGe"].get("diverge_mem_alloc_rs_vertex_t", "cdecl")
    mem_alloc_rs_vertex_t.argtypes = [index_t]
    mem_alloc_rs_vertex_t.restype = POINTER(rs_vertex_t)

    mem_alloc_tu_formfactor_t = _libs["divERGe"].get("diverge_mem_alloc_tu_formfactor_t", "cdecl")
    mem_alloc_tu_formfactor_t.argtypes = [index_t]
    mem_alloc_tu_formfactor_t.restype = POINTER(tu_formfactor_t)

    mem_alloc_complex128_t = _libs["divERGe"].get("diverge_mem_alloc_complex128_t", "cdecl")
    mem_alloc_complex128_t.argtypes = [index_t]
    mem_alloc_complex128_t.restype = POINTER(c_double)

    mem_free = _libs["divERGe"].get("diverge_mem_free", "cdecl")
    mem_free.argtypes = [POINTER(None)]
    mem_free.restype = None

    # Model validation and internals
    model_validate = _libs["divERGe"].get("diverge_model_validate", "cdecl")
    model_validate.argtypes = [POINTER(model_t)]
    model_validate.restype = c_int

    model_internals_common = _libs["divERGe"].get("diverge_model_internals_common", "cdecl")
    model_internals_common.argtypes = [POINTER(model_t)]
    model_internals_common.restype = None

    model_internals_grid = _libs["divERGe"].get("diverge_model_internals_grid", "cdecl")
    model_internals_grid.argtypes = [POINTER(model_t)]
    model_internals_grid.restype = None

    model_internals_patch = _libs["divERGe"].get("diverge_model_internals_patch", "cdecl")
    model_internals_patch.argtypes = [POINTER(model_t), index_t]
    model_internals_patch.restype = None

    model_internals_reset = _libs["divERGe"].get("diverge_model_internals_reset", "cdecl")
    model_internals_reset.argtypes = [POINTER(model_t)]
    model_internals_reset.restype = None

    max_dist_iobi = _libs["divERGe"].get("diverge_max_dist_iobi", "cdecl")
    max_dist_iobi.argtypes = None
    max_dist_iobi.restype = c_double

    model_internals_tu = _libs["divERGe"].get("diverge_model_internals_tu", "cdecl")
    model_internals_tu.argtypes = [POINTER(model_t), c_double]
    model_internals_tu.restype = None

    class ArgUnion(Union):
        _fields_ = [("np_ibz", index_t),
                    ("max_dist", c_double)]
    model_internals_any_ = _libs["divERGe"].get("diverge_model_internals_any", "cdecl")
    model_internals_any_.argtypes = [POINTER(model_t), String, ArgUnion]
    model_internals_any_.restype = None
    # Function: model_internals_any_PY
    # wraps <diverge_model_internals_any> with varargs substituted by kwargs
    # np_ibz or max_dist.
    def model_internals_any_PY( model, mode, np_ibz=None, max_dist=None ):
        arg = ArgUnion()
        if not np_ibz is None:
            arg.np_ibz = np_ibz
        if not max_dist is None:
            arg.max_dist = max_dist
        model_internals_any_( model, mode, arg )

    model_internals_get_E = _libs["divERGe"].get("diverge_model_internals_get_E", "cdecl")
    model_internals_get_E.argtypes = [POINTER(model_t)]
    model_internals_get_E.restype = POINTER(c_double)

    model_internals_get_U = _libs["divERGe"].get("diverge_model_internals_get_U", "cdecl")
    model_internals_get_U.argtypes = [POINTER(model_t)]
    model_internals_get_U.restype = POINTER(complex128_t)

    model_internals_get_H = _libs["divERGe"].get("diverge_model_internals_get_H", "cdecl")
    model_internals_get_H.argtypes = [POINTER(model_t)]
    model_internals_get_H.restype = POINTER(complex128_t)

    model_internals_get_kmesh = _libs["divERGe"].get("diverge_model_internals_get_kmesh", "cdecl")
    model_internals_get_kmesh.argtypes = [POINTER(model_t)]
    model_internals_get_kmesh.restype = POINTER(c_double)

    model_internals_get_kfmesh = _libs["divERGe"].get("diverge_model_internals_get_kfmesh", "cdecl")
    model_internals_get_kfmesh.argtypes = [POINTER(model_t)]
    model_internals_get_kfmesh.restype = POINTER(c_double)

    model_internals_get_greens = _libs["divERGe"].get("diverge_model_internals_get_greens", "cdecl")
    model_internals_get_greens.argtypes = [POINTER(model_t)]
    model_internals_get_greens.restype = POINTER(gf_complex_t)

    # Filling
    model_get_filling = _libs["divERGe"].get("diverge_model_get_filling", "cdecl")
    model_get_filling.argtypes = [POINTER(model_t), POINTER(c_double), index_t]
    model_get_filling.restype = c_double

    model_set_filling = _libs["divERGe"].get("diverge_model_set_filling", "cdecl")
    model_set_filling.argtypes = [POINTER(model_t), POINTER(c_double), index_t, c_double]
    model_set_filling.restype = c_double

    model_set_chempot = _libs["divERGe"].get("diverge_model_set_chempot", "cdecl")
    model_set_chempot.argtypes = [POINTER(model_t), POINTER(c_double), index_t, c_double]
    model_set_chempot.restype = None

    # Flow step
    class struct_flow_step_t(Structure):
        pass
    flow_step_t = struct_flow_step_t

    flow_step_init = _libs["divERGe"].get("diverge_flow_step_init", "cdecl")
    flow_step_init.argtypes = [POINTER(model_t), String, String]
    flow_step_init.restype = POINTER(flow_step_t)

    flow_step_init_any = _libs["divERGe"].get("diverge_flow_step_init_any", "cdecl")
    flow_step_init_any.argtypes = [POINTER(model_t), String]
    flow_step_init_any.restype = POINTER(flow_step_t)

    flow_step_vertmax = _libs["divERGe"].get("diverge_flow_step_vertmax", "cdecl")
    flow_step_vertmax.argtypes = [POINTER(flow_step_t), c_voidp]
    flow_step_vertmax.restype = None

    flow_step_loopmax = _libs["divERGe"].get("diverge_flow_step_loopmax", "cdecl")
    flow_step_loopmax.argtypes = [POINTER(flow_step_t), c_voidp]
    flow_step_loopmax.restype = None

    flow_step_chanmax = _libs["divERGe"].get("diverge_flow_step_chanmax", "cdecl")
    flow_step_chanmax.argtypes = [POINTER(flow_step_t), c_voidp]
    flow_step_chanmax.restype = None

    flow_step_euler = _libs["divERGe"].get("diverge_flow_step_euler", "cdecl")
    flow_step_euler.argtypes = [POINTER(flow_step_t), c_double, c_double]
    flow_step_euler.restype = None

    flow_step_niter = _libs["divERGe"].get("diverge_flow_step_niter", "cdecl")
    flow_step_niter.argtypes = [POINTER(flow_step_t)]
    flow_step_niter.restype = index_t

    flow_step_ntimings = _libs["divERGe"].get("diverge_flow_step_ntimings", "cdecl")
    flow_step_ntimings.argtypes = [POINTER(flow_step_t)]
    flow_step_ntimings.restype = index_t

    flow_step_timing = _libs["divERGe"].get("diverge_flow_step_timing", "cdecl")
    flow_step_timing.argtypes = [POINTER(flow_step_t), index_t]
    flow_step_timing.restype = c_double

    flow_step_timing_descr = _libs["divERGe"].get("diverge_flow_step_timing_descr", "cdecl")
    flow_step_timing_descr.argtypes = [POINTER(flow_step_t), index_t]
    flow_step_timing_descr.restype = c_char_p

    flow_step_lambda = _libs["divERGe"].get("diverge_flow_step_lambda", "cdecl")
    flow_step_lambda.argtypes = [POINTER(flow_step_t)]
    flow_step_lambda.restype = c_double

    flow_step_dlambda = _libs["divERGe"].get("diverge_flow_step_dlambda", "cdecl")
    flow_step_dlambda.argtypes = [POINTER(flow_step_t)]
    flow_step_dlambda.restype = c_double

    flow_step_free = _libs["divERGe"].get("diverge_flow_step_free", "cdecl")
    flow_step_free.argtypes = [POINTER(flow_step_t)]
    flow_step_free.restype = None

    class struct_flow_step_vertex_t(Structure):
        pass
    flow_step_vertex_t = struct_flow_step_vertex_t
    struct_flow_step_vertex_t.__slots__ = [
        'ary',
        'q_0',
        'q_1',
        'nk',
        'n_orbff',
        'n_spin',
        'backend',
        'channel'
    ]
    struct_flow_step_vertex_t._fields_ = [
        ('ary', POINTER(complex128_t)),
        ('q_0', index_t),
        ('q_1', index_t),
        ('nk', index_t),
        ('n_orbff', index_t),
        ('n_spin', index_t),
        ('backend', c_char),
        ('channel', c_char)
    ]
    flow_step_vertex = _libs["divERGe"].get("diverge_flow_step_vertex")
    flow_step_vertex.argtypes = [POINTER(flow_step_t), c_char]
    flow_step_vertex.restype = struct_flow_step_vertex_t

    flow_step_refill = _libs["divERGe"].get("diverge_flow_step_refill")
    flow_step_refill.argtypes = [POINTER(flow_step_t), c_double, c_void_p]
    flow_step_refill.restype = None
    flow_step_refill_Hself = _libs["divERGe"].get("diverge_flow_step_refill_Hself")
    flow_step_refill_Hself.argtypes = [POINTER(flow_step_t), c_double, c_void_p]
    flow_step_refill_Hself.restype = c_double


    # Euler integrator
    class struct_euler_t(Structure):
        pass
    euler_t = struct_euler_t
    struct_euler_t.__slots__ = [
        'Lambda',
        'dLambda',
        'Lambda_min',
        'dLambda_min',
        'dLambda_fac',
        'dLambda_fac_scale',
        'maxvert',
        'maxvert_hard_limit',
        'niter',
        'maxiter',
        'consider_maxvert_iter_start',
        'consider_maxvert_lambda'
    ]
    struct_euler_t._fields_ = [
        ('Lambda', c_double),
        ('dLambda', c_double),
        ('Lambda_min', c_double),
        ('dLambda_min', c_double),
        ('dLambda_fac', c_double),
        ('dLambda_fac_scale', c_double),
        ('maxvert', c_double),
        ('maxvert_hard_limit', c_double),
        ('niter', index_t),
        ('maxiter', index_t),
        ('consider_maxvert_iter_start', index_t),
        ('consider_maxvert_lambda', c_double)
    ]

    euler_defaults_CPP = _libs["divERGe"].get("diverge_euler_defaults_CPP", "cdecl")
    euler_defaults_CPP.argtypes = []
    euler_defaults_CPP.restype = euler_t

    euler_next = _libs["divERGe"].get("diverge_euler_next", "cdecl")
    euler_next.argtypes = [POINTER(euler_t), c_double]
    euler_next.restype = c_bool

    # post processing
    class struct_postprocess_conf_t(Structure):
        pass
    struct_postprocess_conf_t.__slots__ = [
        'patch_q_matrices',
        'patch_q_matrices_use_dV',
        'patch_q_matrices_nv',
        'patch_q_matrices_max_rel',
        'patch_q_matrices_eigen_which',
        'patch_V',
        'patch_dV',
        'patch_Lp',
        'patch_Lm',
        'grid_lingap_vertex_file_P',
        'grid_lingap_vertex_file_C',
        'grid_lingap_vertex_file_D',
        'grid_n_singular_values',
        'grid_use_loop',
        'grid_vertex_file',
        'grid_vertex_chan',
        'tu_which_solver_mode',
        'tu_skip_channel_calc',
        'tu_storing_threshold',
        'tu_storing_relative',
        'tu_n_singular_values',
        'tu_lingap',
        'tu_susceptibilities_full',
        'tu_susceptibilities_ff',
        'tu_selfenergy',
        'tu_channels',
        'tu_symmetry_maps',
    ]
    struct_postprocess_conf_t._fields_ = [
        ('patch_q_matrices', c_bool),
        ('patch_q_matrices_use_dV', c_bool),
        ('patch_q_matrices_nv', c_int),
        ('patch_q_matrices_max_rel', c_double),
        ('patch_q_matrices_eigen_which', c_char),
        ('patch_V', c_bool),
        ('patch_dV', c_bool),
        ('patch_Lp', c_bool),
        ('patch_Lm', c_bool),
        ('grid_lingap_vertex_file_P', c_char * int(1024)),
        ('grid_lingap_vertex_file_C', c_char * int(1024)),
        ('grid_lingap_vertex_file_D', c_char * int(1024)),
        ('grid_n_singular_values', c_int),
        ('grid_use_loop', c_bool),
        ('grid_vertex_file', c_char * int(1024)),
        ('grid_vertex_chan', c_char),
        ('tu_which_solver_mode', c_char),
        ('tu_skip_channel_calc', c_bool),
        ('tu_storing_threshold', c_double),
        ('tu_storing_relative', c_double),
        ('tu_n_singular_values', index_t),
        ('tu_lingap', c_bool),
        ('tu_susceptibilities_full', c_bool),
        ('tu_susceptibilities_ff', c_bool),
        ('tu_selfenergy', c_bool),
        ('tu_channels', c_bool),
        ('tu_symmetry_maps', c_bool),
    ]
    postprocess_conf_t = struct_postprocess_conf_t

    postprocess_conf_defaults_CPP = _libs["divERGe"].get("diverge_postprocess_conf_defaults_CPP", "cdecl")
    postprocess_conf_defaults_CPP.argtypes = []
    postprocess_conf_defaults_CPP.restype = postprocess_conf_t

    postprocess_and_write = _libs["divERGe"].get("diverge_postprocess_and_write", "cdecl")
    postprocess_and_write.argtypes = [POINTER(flow_step_t), String]
    postprocess_and_write.restype = None

    postprocess_and_write_finegrained = _libs["divERGe"].get("diverge_postprocess_and_write_finegrained", "cdecl")
    postprocess_and_write_finegrained.argtypes = [POINTER(flow_step_t), String, POINTER(postprocess_conf_t)]
    postprocess_and_write_finegrained.restype = None

    # Function: postprocess_and_write_PY
    # wraps <diverge_postprocess_and_write_finegrained at
    # diverge_postprocess_conf_t.diverge_postprocess_and_write_finegrained> in a
    # convenient pythonic way.
    #
    # Parameters:
    # step - <diverge_flow_step_t> pointer
    # filename - string: where to save the postprocessing results
    # **kwargs - each member of <diverge_postprocess_conf_t> can be passed as
    #            kwarg (key1 = val1, key2 = val2, …). The defaults are those
    #            returned by <diverge_postprocess_conf_defaults_CPP at
    #            diverge_postprocess_conf_t.diverge_postprocess_conf_defaults_CPP>.
    def postprocess_and_write_PY( step, filename, **kwargs ):
        cfg = postprocess_conf_defaults_CPP()
        for s in struct_postprocess_conf_t.__slots__:
            try:
                v = kwargs.pop(s)
                cfg.__setattr__( s, v )
            except KeyError:
                pass
        for k in kwargs.keys():
            mpi_py_eprint( "postprocess_and_write_PY: unused kwarg '%s'" % k )
        return postprocess_and_write_finegrained( step, filename, byref(cfg) )

    # Model output
    model_output_set_npath = _libs["divERGe"].get("diverge_model_output_set_npath", "cdecl")
    model_output_set_npath.argtypes = [c_int]
    model_output_set_npath.restype = None

    model_to_file = _libs["divERGe"].get("diverge_model_to_file", "cdecl")
    model_to_file.argtypes = [POINTER(model_t), String]
    if sizeof(c_int) == sizeof(c_void_p):
        model_to_file.restype = ReturnString
    else:
        model_to_file.restype = String
        model_to_file.errcheck = ReturnString

    # config
    class struct_model_output_conf_t(Structure):
        pass
    struct_model_output_conf_t.__slots__ = [
        'kc', 'kf', 'kc_ibz_path', 'kf_ibz_path', 'H', 'U', 'E', 'npath', 'fatbands',
    ]
    struct_model_output_conf_t._fields_ = [
        ('kc', c_int),
        ('kf', c_int),
        ('kc_ibz_path', c_int),
        ('kf_ibz_path', c_int),
        ('H', c_int),
        ('U', c_int),
        ('E', c_int),
        ('npath', c_int),
        ('fatbands', c_int),
    ]
    model_output_conf_t = struct_model_output_conf_t

    model_output_conf_defaults_CPP = _libs["divERGe"].get("diverge_model_output_conf_defaults_CPP", "cdecl")
    model_output_conf_defaults_CPP.argtypes = []
    model_output_conf_defaults_CPP.restype = model_output_conf_t

    model_to_file_finegrained = _libs["divERGe"].get("diverge_model_to_file_finegrained", "cdecl")
    model_to_file_finegrained.argtypes = [POINTER(model_t), String, POINTER(model_output_conf_t)]
    if sizeof(c_int) == sizeof(c_void_p):
        model_to_file_finegrained.restype = ReturnString
    else:
        model_to_file_finegrained.restype = String
        model_to_file_finegrained.errcheck = ReturnString

    # Function: model_to_file_PY
    # writes a <diverge_model_t> to file using the
    # <diverge_model_to_file_finegrained at
    # diverge_model_output_conf_t.diverge_model_to_file_finegrained> backend
    # wrapped in a convenient pythonic way. Returns the md5sum of the model file
    # as string.
    #
    # Parameters:
    # model - <diverge_model_t> pointer
    # filename - string: where to save the model
    # **kwargs - each member of <diverge_model_output_conf_t> can be passed as
    #            kwarg (key1 = val1, key2 = val2, …). The defaults are those
    #            returned by <diverge_model_output_conf_defaults_CPP at
    #            diverge_model_output_conf_t.diverge_model_output_conf_defaults_CPP>.
    def model_to_file_PY( model, filename, **kwargs ):
        cfg = model_output_conf_defaults_CPP()
        for s in struct_model_output_conf_t.__slots__:
            try:
                v = kwargs.pop(s)
                cfg.__setattr__( s, v )
            except KeyError:
                pass
        for k in kwargs.keys():
            mpi_py_eprint( "model_to_file_PY: unused kwarg '%s'" % k )
        return model_to_file_finegrained( model, filename, byref(cfg) )

    # Momentum generator
    model_generate_meshes = _libs["divERGe"].get("diverge_model_generate_meshes", "cdecl")
    model_generate_meshes.argtypes = [POINTER(c_double), POINTER(c_double), index_t * int(3), index_t * int(3), (c_double * int(3)) * int(3)]
    model_generate_meshes.restype = None

    model_generate_mom_basis = _libs["divERGe"].get("diverge_model_generate_mom_basis", "cdecl")
    model_generate_mom_basis.argtypes = [(c_double * int(3)) * int(3), (c_double * int(3)) * int(3)]
    model_generate_mom_basis.restype = None

    # Patching
    patching_find_fs_pts_C_ = _libs["divERGe"].get("diverge_patching_find_fs_pts_C", "cdecl")
    patching_find_fs_pts_C_.argtypes = [POINTER(model_t), POINTER(c_double), index_t, index_t, index_t, POINTER(POINTER(index_t)), POINTER(index_t)]
    patching_find_fs_pts_C_.restype = None
    # Function: patching_find_fs_pts_PY
    # returns a numpy array of <index_t> with the indices of the found patches
    #
    # Parameters:
    # model - <diverge_model_t> pointer
    # energies - either None or pointer to a double (64bit) floating point array
    #            that holds energies for nbands bands on the mesh defined in the
    #            model, i.e. an (nk, nb) array
    # nbands - number of bands to use. if energies is None, must coincide with the
    #          number of bands defined in the model
    # np_ibz - number of patches that should be found in the IBZ
    # np_ibz_search - number of points that are considered for the patch search
    def patching_find_fs_pts_PY( model, energies, nbands, np_ibz, np_ibz_search ):
        if energies is None:
            e_ptr = None
        else:
            e_ptr = energies.ctypes.data
        pts = (POINTER(index_t))()
        npts = index_t()
        patching_find_fs_pts_C_( model, e_ptr, nbands, np_ibz,
                    np_ibz_search, ctypes.byref(pts), ctypes.byref(npts) )
        pts_ary = view_array( pts, dtype=index_t, shape=(np.array(npts),) )
        pts_cpy = np.copy( pts_ary )
        mem_free( pts )
        return pts_cpy

    patching_free = _libs["divERGe"].get("diverge_patching_free", "cdecl")
    patching_free.argtypes = [POINTER(mom_patching_t)]
    patching_free.restype = None

    patching_from_indices = _libs["divERGe"].get("diverge_patching_from_indices", "cdecl")
    patching_from_indices.argtypes = [POINTER(model_t), c_voidp, index_t]
    patching_from_indices.restype = POINTER(mom_patching_t)

    patching_autofine = _libs["divERGe"].get("diverge_patching_autofine", "cdecl")
    patching_autofine.argtypes = [POINTER(model_t), POINTER(mom_patching_t), POINTER(c_double), index_t, index_t, c_double, c_double, c_double]
    patching_autofine.restype = None

    patching_symmetrize_refinement = _libs["divERGe"].get("diverge_patching_symmetrize_refinement", "cdecl")
    patching_symmetrize_refinement.argtypes = [POINTER(model_t), POINTER(mom_patching_t)]
    patching_symmetrize_refinement.restype = None

    # Symmetrize
    generate_symm_maps = _libs["divERGe"].get("diverge_generate_symm_maps", "cdecl")
    generate_symm_maps.argtypes = [POINTER(model_t)]
    generate_symm_maps.restype = None

    symmetrize_2pt_coarse = _libs["divERGe"].get("diverge_symmetrize_2pt_coarse", "cdecl")
    symmetrize_2pt_coarse.argtypes = [POINTER(model_t), POINTER(complex128_t), POINTER(complex128_t)]
    symmetrize_2pt_coarse.restype = c_double

    symmetrize_2pt_fine = _libs["divERGe"].get("diverge_symmetrize_2pt_fine", "cdecl")
    symmetrize_2pt_fine.argtypes = [POINTER(model_t), POINTER(complex128_t), POINTER(complex128_t)]
    symmetrize_2pt_fine.restype = c_double

    symmetrize_mom_coarse = _libs["divERGe"].get("diverge_symmetrize_mom_coarse", "cdecl")
    symmetrize_mom_coarse.argtypes = [POINTER(model_t), POINTER(c_double), index_t, POINTER(c_double)]
    symmetrize_mom_coarse.restype = c_double

    symmetrize_mom_fine = _libs["divERGe"].get("diverge_symmetrize_mom_fine", "cdecl")
    symmetrize_mom_fine.argtypes = [POINTER(model_t), POINTER(c_double), index_t, POINTER(c_double)]
    symmetrize_mom_fine.restype = c_double

    # Additional symmetry related stuff
    merge_rs_orb = _libs["divERGe"].get("merge_rs_orb", "cdecl")
    merge_rs_orb.argtypes = [POINTER(model_t)]
    merge_rs_orb.restype = None

    # Testing
    run_tests_ = _libs["divERGe"].get("diverge_run_tests", "cdecl")
    run_tests_.argtypes = [c_int, POINTER(POINTER(c_char))]
    run_tests_.restype = c_int
    # Function: run_tests
    # run all unit tests shipped with divERGe to check health.
    #
    # Parameters:
    # args - list of strings to pass as an argument to CATCH (the testing
    #        framework). To enable a specific test filter, set args=['[filter]'].
    #        For example, the BHK tests can be run with args=['[BHK]'].
    def run_tests( args=[] ):
        try:
            arr = (c_char_p * (len(args) + 1))()
            arr[1:] = [ a.encode('utf8') for a in args ]
            arr[0] = b"diverge.py"
            return run_tests_( len(arr), POINTER(POINTER(c_char))(arr) )
        except:
            mpi_py_eprint("could not find test function")
            return 1

    # hacking
    model_hack = _libs["divERGe"].get("diverge_model_hack", "cdecl")
    model_hack.argtypes = [POINTER(model_t), String, String]
    model_hack.restype = None

    model_print_hacks = _libs["divERGe"].get("diverge_model_print_hacks", "cdecl")
    model_print_hacks.argtypes = None
    model_print_hacks.restype = None

    # shared memory...
    shared_malloc = _libs['divERGe'].get('shared_malloc', 'cdecl')
    shared_malloc.argtypes = [c_int64]
    shared_malloc.restype = c_voidp
    shared_calloc = _libs['divERGe'].get('shared_calloc', 'cdecl')
    shared_calloc.argtypes = [c_int64, c_int64]
    shared_calloc.restype = c_voidp
    shared_free = _libs['divERGe'].get('shared_free', 'cdecl')
    shared_free.argtypes = [c_voidp]
    shared_free.restype = None
    shared_exclusive_enter = _libs['divERGe'].get('shared_exclusive_enter', 'cdecl')
    shared_exclusive_enter.argtypes = [c_voidp]
    shared_exclusive_enter.restype = c_int
    shared_exclusive_wait = _libs['divERGe'].get('shared_exclusive_wait', 'cdecl')
    shared_exclusive_wait.argtypes = [c_voidp]
    shared_exclusive_wait.restype = None
    shared_malloc_rank = _libs['divERGe'].get('shared_malloc_rank', 'cdecl')
    shared_malloc_rank.argtypes = []
    shared_malloc_rank.restype = c_int
    shared_malloc_size = _libs['divERGe'].get('shared_malloc_size', 'cdecl')
    shared_malloc_size.argtypes = []
    shared_malloc_size.restype = c_int
    shared_malloc_barrier = _libs['divERGe'].get('shared_malloc_barrier', 'cdecl')
    shared_malloc_barrier.argtypes = []
    shared_malloc_barrier.restype = None

    # redo the typedefs
    rs_hopping_t = struct_rs_hopping_t
    rs_vertex_t = struct_rs_vertex_t
    mom_patching_t = struct_mom_patching_t
    model_t = struct_model_t
    internals_t = struct_internals_t
    tu_formfactor_t = struct_tu_formfactor_t
    flow_step_t = struct_flow_step_t
    euler_t = struct_euler_t

    # Var: numpy
    # imports numpy in standard fashion (as np)
    import numpy as np
    # Var: zeros
    # additionally import zeros from numpy for less namespace pollution
    zeros = np.zeros

    # Function: view_array
    # give an array view on existing memory
    #
    # Parameters:
    #   dtype - data type that the view is using
    #   shape - all dimensions (in C ordering) that the view is using
    def view_array( mem, dtype=np.complex128, shape=(1,) ):
        dtype = np.dtype(dtype)
        return np.ctypeslib.as_array( cast(mem, POINTER(c_char)), shape=(*shape,dtype.itemsize) ).view( dtype=dtype ).reshape( shape )

    # Function: alloc_array
    # allocate an array of given data type and shape such that it is not cleared by
    # the python garbage collector
    #
    # Parameters:
    #   shape - all dimensions (in C ordering)
    #   dtype - data type. can be <complex128_t>, <rs_hopping_t>, <rs_vertex_t>, or
    #           <tu_formfactor_t>. data type name must be passed as python string.
    def alloc_array( shape, dtype = "complex128_t" ):
        if dtype == "complex128_t":
            return view_array( mem_alloc_complex128_t(np.prod(shape)), dtype=np.complex128, shape=shape )
        elif dtype == "rs_hopping_t":
            return view_array( mem_alloc_rs_hopping_t(np.prod(shape)), dtype=struct_rs_hopping_t, shape=shape )
        elif dtype == "rs_vertex_t":
            return view_array( mem_alloc_rs_vertex_t(np.prod(shape)), dtype=struct_rs_vertex_t, shape=shape )
        elif dtype == "tu_formfactor_t":
            return view_array( mem_alloc_tu_formfactor_t(np.prod(shape)), dtype=struct_tu_formfactor_t, shape=shape )
        else:
            mpi_py_eprint("cannot allocate array of type '%s'" % dtype)
            return None

    print_ = print
    # Function: print
    # wraps the python print function to work with MPI
    def print( *args, **kwargs ):
        if mpi_comm_rank() == 0:
            print_( *args, **kwargs )

    # Function: autoflow
    # simplest possible FRG flow
    #
    # Parameters:
    #   hoppings - numpy array of dtype <rs_hopping_t>, *required*.
    #   vertex - numpy array of dtype <rs_vertex_t>, *required*.
    #   nk - number of coarse k points in all three dimensions, *required*.
    #        Dimensionality of the model is inferred from the nk list by inspecting
    #        how many nonzero entries exist.
    #   nkf - number of fine k points in all three dimensions, defaults to 15 in
    #         each dimension for mode=="grid" or mode=="tu", and to 1 for
    #         mode=="patch".
    #   lattice - 3×3 matrix in C order that holds the lattice vectors lattice[0],
    #             lattice[1], and lattice[2]
    #   no - number of orbitals and sites
    #   SU2 - SU2 symmetry setting. If SU2>0, model is assumed to be SU2 symmetric.
    #         If SU2==0, model is not SU2 symmetric and exchange symmetry in the
    #         vertex is *not* enforced. for SU2<0, model is not SU2 symmetric, but
    #         exchange symmetry in the vertex *is* enforced.
    #   nspin - number of spin indices. For SU2>0, nspin=1 is required. For SU2<=0,
    #           nspin must be set from physical grounds.
    #   sites - positions of the orbital centers in real space as array of shape
    #           (no, 3).
    #   model_output - output file for the model.
    #   post_output - output file for postprocessing info.
    #   flow_output - output file for flow data.
    #   rs_symmetries - realspace symmetry list as array of shape (nsym, 3, 3). If
    #                   None, no symmetries are used.
    #   orb_symmetries - orbital symmetry list as array of shape (nsym, no, no). If
    #                    None, no symmetries are used.
    #   mode - backend selection. valid choices are "grid", "patch", and "tu".
    #   npatches_ibz - number of FS patches in the IBZ used with the "patch"
    #                  backend.
    #   formfactor_maxdist - maximal distance up to which formfactors should be
    #                        included in the "tu" backend.
    #   channels - selection of channels to be included in the FRG flow. The library
    #              looks for the characters 'P', 'C', 'D', 'S' in the string that
    #              stand for the particle particle channel ('P'), the crossed
    #              particle hole channel ('C'), the direct particle hole channel
    #              ('D'), and self-energies ('S'), respectively.
    #   ibz_path - Brillouin zone path in reciprocal crystal coordinates. if empty,
    #              do not compute band structure
    #   maxvert - value after which the vertex is considered 'diverged'
    #   mu - chemical potential. use if mu is not None
    #   nu - filling value. use if nu is not None. nu has precedence over mu, i.e.
    #        if both nu and mu are not None, the filling value is used.
    #
    # Create the (numpy) arrays with
    # === Python ===
    # import diverge as dvg
    # hoppings = dvg.zeros(shape_h, dtype=dvg.rs_hopping_t)
    # vertex = dvg.zeros(shape_v, dtype=dvg.rs_vertex_t)
    # ==============
    def autoflow( hoppings=None, vertex=None, nk=None, nkf=None, lattice=np.eye(3),
                 no=1, SU2=1, nspin=1, sites=[[0,0,0]], model_output="model.dvg",
                 post_output="post.dvg", flow_output="flow.dat", rs_symmetries=None,
                 orb_symmetries=None, mode="tu", npatches_ibz=6,
                 formfactor_maxdist=2.01, channels="PCD", ibz_path=[], maxvert=50.0,
                 mu=None, nu=None ):

        errors = 0
        init(None, None)
        compilation_status()

        if (hoppings is None) or (vertex is None) or (nk is None):
            mpi_py_eprint("must provide hoppings, vertex, and nk")
            errors += 1

        if len(sites) != no:
            mpi_py_eprint("sites array must be of shape (%i, 3)" % no)
            errors += 1

        if (not (rs_symmetries is None)) and (not (orb_symmetries is None)):
            num_sym = len(rs_symmetries)
            if num_sym != len(orb_symmetries):
                mpi_py_eprint("number or orbital symmetries does not match number of realspace symmetries")
                errors += 1
            if rs_symmetries.shape != (num_sym, 3, 3):
                mpi_py_eprint("realspace symmetries must be of shape (%i, 3, 3)" % num_sym)
                errors += 1
            if orb_symmetries.shape != (num_sym, no, no):
                mpi_py_eprint("orbital symmetries must be of shape (%i, %i, %i)" % (num_sym, no, no))
                errors += 1

        if errors > 0:
            mpi_py_eprint("aborting due to previous errors")
            mpi_exit(errors)

        if SU2>0 and nspin != 1:
            nspin = 1
            mpi_py_eprint("for SU2>0, nspin=1 must be set; resetting")

        if nkf is None:
            nkf = np.copy(nk)
            default_fac = 1 if mode == "patch" else 15
            nkf[nkf != 0] = default_fac
            mpi_py_eprint("using refinement of factor %i as default" % default_fac)

        mean_hop = np.abs(np.array(hoppings['t']).view(np.complex128)).sum()
        mean_vert = np.abs(np.array(vertex['V']).view(np.complex128)).sum()
        if mean_hop < 0.5 or mean_hop > 50:
            mpi_py_eprint("consider readjusting your units, autoflow is optimized for hoppings of order one.")
        if mean_vert < 0.5 or mean_vert > 50:
            mpi_py_eprint("consider readjusting your units, autoflow is optimized for vertices of order one.")

        model = model_init()

        model.contents.name = b"autoflow"

        model.contents.nk[:] = nk
        model.contents.nkf[:] = nkf

        lattice_view = view_array( model.contents.lattice, dtype=np.float64, shape=(3,3) )
        lattice_view[:,:] = lattice

        model.contents.n_orb = no
        positions_view = view_array( model.contents.positions, dtype=np.float64, shape=(no,3) )
        positions_view[:,:] = sites

        model.contents.SU2 = SU2
        model.contents.n_spin = nspin

        hoppings_copy = alloc_array( hoppings.shape, "rs_hopping_t" )
        hoppings_copy[:] = hoppings
        model.contents.hop = hoppings_copy.ctypes.data
        model.contents.n_hop = hoppings_copy.size

        vertex_copy = alloc_array( vertex.shape, "rs_vertex_t" )
        vertex_copy[:] = vertex
        model.contents.vert = vertex_copy.ctypes.data
        model.contents.n_vert = vertex_copy.size

        if not orb_symmetries is None and not rs_symmetries is None:
            mpi_py_eprint( "setting symmetries" )
            orbsym_copy = alloc_array( orb_symmetries.shape, "complex128_t" )
            orbsym_copy[:] = orb_symmetries
            model.contents.orb_symmetries = orbsym_copy.ctypes.data
            rssym_view = view_array( model.contents.rs_symmetries, dtype=np.float64, shape=rs_symmetries.shape )
            rssym_view[:] = rs_symmetries
            model.contents.n_sym = orbsym_copy.shape[0]

        if len(ibz_path) > 0:
            model.contents.n_ibz_path = len(ibz_path)
            ibz_view = view_array( model.contents.ibz_path, dtype=np.float64, shape=(len(ibz_path), 3) )
            ibz_view[:] = ibz_path

        validate = model_validate( model )
        if validate:
            mpi_py_eprint("invalid model")
            mpi_exit( validate )

        model_internals_common( model )

        set_fill = False
        if not mu is None:
            model_set_chempot( model, None, -1, mu )
            set_fill = True
        if not nu is None:
            model_set_filling( model, None, -1, nu )
            set_fill = True
        if set_fill:
            mpi_py_eprint( "adjusted mu/nu. filling value: %.3f" % model_get_filling( model, None, -1 ) )

        if mode == "grid":
            model_internals_grid( model )
        elif mode == "patch":
            model_internals_patch( model, npatches_ibz )
        elif mode == "tu":
            model_internals_tu( model, formfactor_maxdist )
        checksum = model_to_file( model, model_output )
        mpi_py_eprint( "model output to %s, checksum %s" % (model_output, checksum) )

        step = flow_step_init( model, mode, channels )
        maxs = dict(vert=np.zeros(1), loop=np.zeros(2), chan=np.zeros(3))
        eu = euler_defaults_CPP()
        eu.dLambda_fac_scale = 1.0
        eu.dLambda_fac = 0.1
        eu.maxvert = maxvert

        mpi_py_eprint( "flow output to %s…" % flow_output )
        stdout_ = sys.stdout
        sys.stdout = mpi_stdout_logger( flow_output )

        print( "# %9s %12s %11s %11s %11s %11s %11s %11s" % ('Lambda', 'dLambda', 'Lp', 'Lm', 'dP', 'dC', 'dD', 'V') )
        while True:
            flow_step_euler( step, eu.Lambda, eu.dLambda )
            flow_step_vertmax( step, maxs['vert'].ctypes.data )
            flow_step_loopmax( step, maxs['loop'].ctypes.data )
            flow_step_chanmax( step, maxs['chan'].ctypes.data )
            print( "%.5e %.5e %.5e %.5e %.5e %.5e %.5e %.5e" %
                  (eu.Lambda, eu.dLambda, *maxs['loop'], *maxs['chan'], *maxs['vert']) )
            eu_next = euler_next( byref(eu), maxs['vert'][0] )
            if not eu_next:
                break
            sys.stdout.flush()

        sys.stdout = stdout_

        mpi_py_eprint( "postprocessing output to %s…" % post_output )
        postprocess_and_write( step, post_output )

        flow_step_free( step )
        model_free( model )
        finalize()

    # Class: mpi_stdout_logger
    # direct output to stdout and a file at the same time. usage:
    # === Python ===
    # stdout_ = sys.stdout
    # sys.stdout = mpi_stdout_logger(filename)
    # # some code with print() statements
    # sys.stdout = stdout_
    # ==============
    class mpi_stdout_logger(object):
        def __init__(self, fname, mode="w"):
            if mpi_comm_rank() == 0:
                self.terminal = sys.stdout
                self.log = open(fname, mode)
        def write(self, message):
            if mpi_comm_rank() == 0:
                self.terminal.write(message)
                self.log.write(message)
        def flush(self):
            if mpi_comm_rank() == 0:
                self.terminal.flush()
                self.log.flush()

else: # _libs["divERGe"] is None
    print( info() )

# Namespace: Structured Arrays
# DivERGe defines several primitive datatypes (structures) that can be
# interfaced to from this python library:
#
# - <complex128_t>
# - <rs_hopping_t>
# - <rs_vertex_t>
# - <tu_formfactor_t>
# - <sym_op_t>
# - <site_descr_t>
#
# Allocation:
# To allocate an array that is *not* garbage collected from python (and can thus
# be free'd from within divERGe or via <diverge.mem_free at diverge_mem_free>),
# call <diverge.alloc_array at dvg.alloc_array>. The datatype is passed as
# *string* to this function, in contrast to what one may be used to from numpy.
# For standard, garbage-collected allocation, use <diverge.zeros at dvg.zeros>
# (or just np.zeros):
# === Python ===
# hopping_gc = diverge.zeros( (128,), dtype=diverge.rs_hopping_t )
# hopping_non_gc = diverge.alloc_array( (128,), dtype="rs_hopping_t" )
# ==============
#
# View:
# A numpy structured array can be used just as any other numpy array (shapes,
# views, etc). To conveniently use memory owned by divERGe in a pythonic way, we
# added <diverge.view_array at dvg.view_array>:
# === Python ===
# positions = diverge.view_array( model.contents.positions, dtype=np.float64, shape=(12,3) )
# positions[:] += 1 # now positions are updated!
# ==============
#
# Access:
# Now to the usage of structured arrays (or array views). If one or more full
# element(s) is (are) to be set, a (list of) tuple(s) can be passed as so:
# === Python ===
# hopping = diverge.alloc_array( (128,), dtype="rs_hopping_t" )
# hopping[0] = ( [0,0,0], 0,0, 0,0, (1.0,0.0) )
# hopping[1:3] = [ ( [0,0,0], 1,1, 0,0, (0.2,0.0) ),
#                  ( [0,0,0], 2,2, 0,0, (0.2,0.0) ) ]
# ==============
# This works because the inserted tuples are understood as compatible with the
# <rs_hopping_t> data type:
# ---- C/C++ ----
# // C declaration
# struct rs_hopping_t {
#   index_t R[3];
#   index_t o1,o2, s1,s2;
#   complex128_t t;
# };
# ---------------
# If only a single member of the structure should be modified or accessed, a
# dictionary-like syntax is supported:
# === Python ===
# vertex = diverge.alloc_array( (128,), dtype="rs_vertex_t" )
# vertex[0]['chan'] = 'D'
# vertex[0]['V'] = (3.0,0.0)
# ==============
# The keys of this dictionary-like access resemble the names of the members in
# the respective C declaration.

